import pytest
from _pytest.fixtures import FixtureRequest

from metricflow.dataflow.sql_table import SqlTable
from metricflow.sql.optimizer.column_pruner import SqlColumnPrunerOptimizer
from metricflow.sql.render.sql_plan_renderer import DefaultSqlQueryPlanRenderer, SqlQueryPlanRenderer
from metricflow.sql.sql_exprs import (
    SqlStringExpression,
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlComparisonExpression,
    SqlComparison,
    SqlIsNullExpression,
)
from metricflow.sql.sql_plan import (
    SqlSelectColumn,
    SqlTableFromClauseNode,
    SqlSelectStatementNode,
    SqlJoinDescription,
    SqlJoinType,
    SqlQueryPlanNode,
)
from metricflow.test.fixtures.setup_fixtures import MetricFlowTestSessionState
from metricflow.test.sql.compare_sql_plan import assert_default_rendered_sql_equal


@pytest.fixture
def column_pruner() -> SqlColumnPrunerOptimizer:  # noqa: D
    return SqlColumnPrunerOptimizer()


@pytest.fixture
def sql_plan_renderer() -> SqlQueryPlanRenderer:  # noqa: D
    return DefaultSqlQueryPlanRenderer()


@pytest.fixture
def base_select_statement() -> SqlSelectStatementNode:
    """SELECT statement used to build test cases.

    In the initial case, all columns are selected from both from_source and joined_source. Test cases will generally
    change the selected columns to see if the sources are pruned appropriately.

    -- test0
    SELECT
      from_source.col0 AS from_source_col0
      , from_source.col1 AS from_source_col1
      , from_source.join_col AS from_source_join_col
      , joined_source.col0 AS joined_source_col0
      , joined_source.col1 AS joined_source_col1
      , joined_source.join_col AS joined_source_join_col
    FROM (
      -- from_source
      SELECT
        from_source_table.col0 AS col0
        , from_source_table.col1 AS col1
        , from_source_table.join_col AS join_col
      FROM demo.from_source_table from_source_table
    ) from_source
    JOIN (
      -- joined_source
      SELECT
        joined_source_table.col0 AS col0
        , joined_source_table.col1 AS col1
        , joined_source_table.join_col AS join_col
      FROM demo.joined_source_table joined_source_table
    ) joined_source
    ON
      from_source.join_col = joined_source.join_col
    """
    return SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col0")
                ),
                column_alias="from_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col1")
                ),
                column_alias="from_source_col1",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="join_col")
                ),
                column_alias="from_source_join_col",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col0")
                ),
                column_alias="joined_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col1")
                ),
                column_alias="joined_source_col1",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="join_col")
                ),
                column_alias="joined_source_join_col",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="from_source",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="col0",
                        )
                    ),
                    column_alias="col0",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="col1",
                        )
                    ),
                    column_alias="col1",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="join_col",
                        )
                    ),
                    column_alias="join_col",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="from_source_table")),
            from_source_alias="from_source_table",
            joins_descs=(),
            group_bys=(),
            order_bys=(),
        ),
        from_source_alias="from_source",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="joined_source",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="col0",
                                )
                            ),
                            column_alias="col0",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="col1",
                                )
                            ),
                            column_alias="col1",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="join_col",
                                )
                            ),
                            column_alias="join_col",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="joined_source_table")
                    ),
                    from_source_alias="joined_source_table",
                    joins_descs=(),
                    group_bys=(),
                    order_bys=(),
                ),
                right_source_alias="joined_source",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="from_source", column_name="join_col")
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="joined_source", column_name="join_col")
                    ),
                ),
                join_type=SqlJoinType.INNER,
            ),
        ),
        where=None,
        group_bys=(),
        order_bys=(),
    )


def test_no_pruning(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where no pruning should occur"""
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=base_select_statement,
        plan_id="before_pruning",
    )
    column_pruned_select_node = column_pruner.optimize(base_select_statement)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


def test_prune_from_source(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where columns should be pruned from the FROM clause"""

    select_statement_with_some_from_source_column_removed = SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col0")
                ),
                column_alias="from_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col0")
                ),
                column_alias="joined_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col1")
                ),
                column_alias="joined_source_col1",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="join_col")
                ),
                column_alias="joined_source_join_col",
            ),
        ),
        from_source=base_select_statement.from_source,
        from_source_alias=base_select_statement.from_source_alias,
        joins_descs=base_select_statement.join_descs,
        group_bys=base_select_statement.group_bys,
        order_bys=base_select_statement.order_bys,
        where=base_select_statement.where,
    )

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=select_statement_with_some_from_source_column_removed,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(select_statement_with_some_from_source_column_removed)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


def test_prune_joined_source(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where columns should be pruned from the JOIN clause"""

    select_statement_with_some_joined_source_column_removed = SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col0")
                ),
                column_alias="from_source_col0",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col1")
                ),
                column_alias="from_source_col1",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="join_col")
                ),
                column_alias="from_source_join_col",
            ),
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="joined_source", column_name="col0")
                ),
                column_alias="joined_source_col0",
            ),
        ),
        from_source=base_select_statement.from_source,
        from_source_alias=base_select_statement.from_source_alias,
        joins_descs=base_select_statement.join_descs,
        group_bys=base_select_statement.group_bys,
        order_bys=base_select_statement.order_bys,
        where=base_select_statement.where,
    )

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=select_statement_with_some_joined_source_column_removed,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(select_statement_with_some_joined_source_column_removed)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


def test_dont_prune_if_in_where(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests that columns aren't pruned from parent sources if columns are used in a where."""

    select_statement_with_other_exprs = SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlColumnReferenceExpression(
                    col_ref=SqlColumnReference(table_alias="from_source", column_name="col0")
                ),
                column_alias="from_source_col0",
            ),
        ),
        from_source=base_select_statement.from_source,
        from_source_alias=base_select_statement.from_source_alias,
        joins_descs=base_select_statement.join_descs,
        where=SqlIsNullExpression(
            SqlColumnReferenceExpression(col_ref=SqlColumnReference(table_alias="from_source", column_name="col1"))
        ),
        group_bys=base_select_statement.group_bys,
        order_bys=base_select_statement.order_bys,
    )

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=select_statement_with_other_exprs,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(select_statement_with_other_exprs)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


def test_dont_prune_with_str_expr(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests that columns aren't pruned from parent sources if there's a string expression in the select."""

    select_statement_with_other_exprs = SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlStringExpression("from_source.col0", requires_parenthesis=False),
                column_alias="some_string_expr",
            ),
        ),
        from_source=base_select_statement.from_source,
        from_source_alias=base_select_statement.from_source_alias,
        joins_descs=base_select_statement.join_descs,
        where=base_select_statement.where,
        group_bys=base_select_statement.group_bys,
        order_bys=base_select_statement.order_bys,
    )

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=select_statement_with_other_exprs,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(select_statement_with_other_exprs)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


def test_prune_with_str_expr(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    base_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests that columns are from parent sources if there's a string expression in the select with known cols."""

    select_statement_with_other_exprs = SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlStringExpression("from_source.col0", requires_parenthesis=False, used_columns=("col0",)),
                column_alias="some_string_expr",
            ),
        ),
        from_source=base_select_statement.from_source,
        from_source_alias=base_select_statement.from_source_alias,
        joins_descs=base_select_statement.join_descs,
        where=base_select_statement.where,
        group_bys=base_select_statement.group_bys,
        order_bys=base_select_statement.order_bys,
    )

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=select_statement_with_other_exprs,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(select_statement_with_other_exprs)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


@pytest.fixture
def string_select_statement() -> SqlSelectStatementNode:
    """Test case for propagating a column used in a string expression.

    -- test0
    SELECT
      col0 AS from_source_col0
    FROM (
      -- from_source
      SELECT
        from_source_table.col0 AS col0
        , from_source_table.col1 AS col1
        , from_source_table.join_col AS join_col
      FROM demo.from_source_table from_source_table
    ) from_source
    JOIN (
      -- joined_source
      SELECT
        joined_source_table.col2 AS col2
        , joined_source_table.col3 AS col3
        , joined_source_table.join_col AS join_col
      FROM demo.joined_source_table joined_source_table
    ) joined_source
    ON
      from_source.join_col = joined_source.join_col
    """
    return SqlSelectStatementNode(
        description="test0",
        select_columns=(
            SqlSelectColumn(
                expr=SqlStringExpression(sql_expr="col0", used_columns=("col0",)),
                column_alias="from_source_col0",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="from_source",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="col0",
                        )
                    ),
                    column_alias="col0",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="col1",
                        )
                    ),
                    column_alias="col1",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="from_source_table",
                            column_name="join_col",
                        )
                    ),
                    column_alias="join_col",
                ),
            ),
            from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="from_source_table")),
            from_source_alias="from_source_table",
            joins_descs=(),
            group_bys=(),
            order_bys=(),
        ),
        from_source_alias="from_source",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="joined_source",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="col2",
                                )
                            ),
                            column_alias="col0",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="col3",
                                )
                            ),
                            column_alias="col1",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="joined_source_table",
                                    column_name="join_col",
                                )
                            ),
                            column_alias="join_col",
                        ),
                    ),
                    from_source=SqlTableFromClauseNode(
                        sql_table=SqlTable(schema_name="demo", table_name="joined_source_table")
                    ),
                    from_source_alias="joined_source_table",
                    joins_descs=(),
                    group_bys=(),
                    order_bys=(),
                ),
                right_source_alias="joined_source",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="from_source", column_name="join_col")
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="joined_source", column_name="join_col")
                    ),
                ),
                join_type=SqlJoinType.INNER,
            ),
        ),
        where=None,
        group_bys=(),
        order_bys=(),
    )


def test_prune_str_expr(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    string_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests a case where a string expr in a node results in the parent being pruned properly"""

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=string_select_statement,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(string_select_statement)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


@pytest.fixture
def grandparent_pruning_select_statement() -> SqlSelectStatementNode:
    """Test case testing pruning of grandparents in a select statement.

    The string expression for col0 does not have "used_columns" set, so the pruning path is different from the previous
    case.

    -- src2
    SELECT
      col0 AS col0
    FROM (
      -- src1
      SELECT
        src1.col0 AS col0
        , src1.col1 AS col1
      FROM (
        -- from_source1
        SELECT
          src0.col0 AS col0
          , src0.col1 AS col1
          , src0.col2 AS col2
        FROM demo.src0 src0
      ) src1
    ) src2
    """
    return SqlSelectStatementNode(
        description="src2",
        select_columns=(
            SqlSelectColumn(
                expr=SqlStringExpression(sql_expr="col0"),
                column_alias="col0",
            ),
        ),
        from_source=SqlSelectStatementNode(
            description="src1",
            select_columns=(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="src1",
                            column_name="col0",
                        )
                    ),
                    column_alias="col0",
                ),
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(
                            table_alias="src1",
                            column_name="col1",
                        )
                    ),
                    column_alias="col1",
                ),
            ),
            from_source=SqlSelectStatementNode(
                description="src0",
                select_columns=(
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="col0",
                            )
                        ),
                        column_alias="col0",
                    ),
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="col1",
                            )
                        ),
                        column_alias="col1",
                    ),
                    SqlSelectColumn(
                        expr=SqlColumnReferenceExpression(
                            col_ref=SqlColumnReference(
                                table_alias="src0",
                                column_name="col2",
                            )
                        ),
                        column_alias="col2",
                    ),
                ),
                from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="src0")),
                from_source_alias="src0",
                joins_descs=(),
                group_bys=(),
                order_bys=(),
            ),
            from_source_alias="src1",
            joins_descs=(),
            group_bys=(),
            order_bys=(),
        ),
        from_source_alias="src2",
        joins_descs=(),
        where=None,
        group_bys=(),
        order_bys=(),
    )


def test_prune_grandparents(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    grandparent_pruning_select_statement: SqlQueryPlanNode,
) -> None:
    """Tests a case where a string expr in a node prevents the parent from being pruned, but prunes grandparents."""

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=grandparent_pruning_select_statement,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(grandparent_pruning_select_statement)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )


@pytest.fixture
def join_grandparent_pruning_select_statement() -> SqlSelectStatementNode:
    """SELECT statement used to build test cases.

    In the initial case, all columns are selected from both from_source and joined_source. Test cases will generally
    change the selected columns to see if the sources are pruned appropriately.

    -- src4
    SELECT
      col0 AS col0
    FROM src3
    JOIN (
      -- joined_source
      SELECT
        src1.col0 AS col0
        sc1.join_col AS join_col
      FROM (
        SELECT
          src0.col0 AS col0
          , src0.col1 AS col1
          , src0.join_col AS join_col
        FROM src0
      ) src1
    ) src4
    ON
      src3.join_col = src4.join_col
    """
    return SqlSelectStatementNode(
        description="4",
        select_columns=(
            SqlSelectColumn(
                expr=SqlStringExpression(sql_expr="col0"),
                column_alias="col0",
            ),
        ),
        from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="from_source_table")),
        from_source_alias="src3",
        joins_descs=(
            SqlJoinDescription(
                right_source=SqlSelectStatementNode(
                    description="src1",
                    select_columns=(
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="src1",
                                    column_name="col0",
                                )
                            ),
                            column_alias="col0",
                        ),
                        SqlSelectColumn(
                            expr=SqlColumnReferenceExpression(
                                col_ref=SqlColumnReference(
                                    table_alias="src1",
                                    column_name="join_col",
                                )
                            ),
                            column_alias="join_col",
                        ),
                    ),
                    from_source=SqlSelectStatementNode(
                        description="src0",
                        select_columns=(
                            SqlSelectColumn(
                                expr=SqlColumnReferenceExpression(
                                    col_ref=SqlColumnReference(
                                        table_alias="src0",
                                        column_name="col0",
                                    )
                                ),
                                column_alias="col0",
                            ),
                            SqlSelectColumn(
                                expr=SqlColumnReferenceExpression(
                                    col_ref=SqlColumnReference(
                                        table_alias="src0",
                                        column_name="col1",
                                    )
                                ),
                                column_alias="col1",
                            ),
                            SqlSelectColumn(
                                expr=SqlColumnReferenceExpression(
                                    col_ref=SqlColumnReference(
                                        table_alias="src0",
                                        column_name="join_col",
                                    )
                                ),
                                column_alias="join_col",
                            ),
                        ),
                        from_source=SqlTableFromClauseNode(sql_table=SqlTable(schema_name="demo", table_name="src0")),
                        from_source_alias="src0",
                        joins_descs=(),
                        group_bys=(),
                        order_bys=(),
                    ),
                    from_source_alias="src1",
                    joins_descs=(),
                    group_bys=(),
                    order_bys=(),
                ),
                right_source_alias="src4",
                on_condition=SqlComparisonExpression(
                    left_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="src3", column_name="join_col")
                    ),
                    comparison=SqlComparison.EQUALS,
                    right_expr=SqlColumnReferenceExpression(
                        col_ref=SqlColumnReference(table_alias="src4", column_name="join_col")
                    ),
                ),
                join_type=SqlJoinType.INNER,
            ),
        ),
        where=None,
        group_bys=(),
        order_bys=(),
    )


def test_prune_grandparents_in_join_query(
    request: FixtureRequest,
    mf_test_session_state: MetricFlowTestSessionState,
    column_pruner: SqlColumnPrunerOptimizer,
    join_grandparent_pruning_select_statement: SqlSelectStatementNode,
) -> None:
    """Tests pruning grandparents of a join query."""

    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=join_grandparent_pruning_select_statement,
        plan_id="before_pruning",
    )

    column_pruned_select_node = column_pruner.optimize(join_grandparent_pruning_select_statement)
    assert_default_rendered_sql_equal(
        request=request,
        mf_test_session_state=mf_test_session_state,
        sql_plan_node=column_pruned_select_node,
        plan_id="after_pruning",
    )



from __future__ import annotations

import dataclasses
import datetime
import logging
from builtins import NameError
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Optional, Type, TypeVar, Tuple, Union, get_args, get_origin, get_type_hints
from typing_extensions import TypeAlias

import pydantic
from pydantic import BaseModel

from metricflow.object_utils import pformat_big_objects

logger = logging.getLogger(__name__)


# Any Pydantic object
PydanticT = TypeVar("PydanticT", bound=Type[BaseModel])
# Any value
AnyValueType: TypeAlias = Any  # type: ignore[misc]


class UnknownClassError(Exception):
    """Raised when there's an issue getting type information for a SerializableDataclass"""

    pass


class DataclassDeserializationError(Exception):
    """Raised when there is any error deserializing a dataclass."""

    pass


def _get_dataclass_field_definitions(dataclass_type: Type) -> Dict[str, FieldDefinition]:
    """Returns the types of fields in a dataclass. Returns a dict from the name of the field to the type."""
    assert dataclasses.is_dataclass(dataclass_type)
    try:
        type_hints = get_type_hints(dataclass_type, localns={})
        fields = dataclasses.fields(dataclass_type)
        return {
            field.name: FieldDefinition(field_type=type_hints[field.name], default_value=field.default)
            for field in fields
        }

    except NameError as e:
        raise UnknownClassError(
            f"Error getting type hints for dataclass {dataclass_type}. Please see the nested exception as a required "
            f"class may not be imported properly."
        ) from e


def _is_optional_type(type_to_check: Type) -> bool:
    """Returns true if the given type is an optional type. Python represents optional as Union[SomeType, NoneType]."""
    if get_origin(type_to_check) is Union:
        args = get_args(type_to_check)
        if len(args) == 2 and issubclass(args[1], type(None)):
            return True
    return False


def _get_type_parameter_for_optional(type_to_check: Type) -> Type:
    """Given Union[SomeType, NoneType], return SomeType."""
    assert _is_optional_type(type_to_check)
    return get_args(type_to_check)[0]


def _is_supported_field_type_in_serializable_dataclass(field_type: Type) -> bool:
    """Returns if a type of a given field in a SerializableDatclass is supported.

    For container classes, this does not check the type of the type parameter for the container.
    """
    return (
        (
            _is_sequence_like_tuple_type(field_type)
            and _is_supported_field_type_in_serializable_dataclass(
                _get_type_parameter_for_sequence_like_tuple_type(field_type)
            )
        )
        or (
            _is_optional_type(field_type)
            and _is_supported_field_type_in_serializable_dataclass(_get_type_parameter_for_optional(field_type))
        )
        or issubclass(field_type, Enum)
        or issubclass(field_type, SerializableDataclass)
        or issubclass(field_type, int)
        or issubclass(field_type, float)
        or issubclass(field_type, str)
        or issubclass(field_type, datetime.datetime)
        or issubclass(field_type, datetime.timedelta)
        or issubclass(field_type, BaseModel)
    )


def _is_sequence_like_tuple_type(field_type: Type) -> bool:
    """Returns true for tuple types that are like sequences.

    In a dataclass definition:

        foo: Tuple[SomeType, ...]

    would return true, but

       foo: Tuple[SomeType, SomeOtherType]

    would not.
    """
    if get_origin(field_type) is tuple:
        args = get_args(field_type)
        return len(args) == 2 and args[1] is Ellipsis
    return False


def _get_type_parameter_for_sequence_like_tuple_type(field_type: Type) -> Type:
    """Return the type parameter for a sequence like tuple type in a daclass definition.

    e.g.

        foo: Tuple[SomeType, ...]

    should return SomeType.
    """
    assert _is_sequence_like_tuple_type(field_type)
    args = get_args(field_type)
    return args[0]


class SerializableDataclass:
    """Describes a dataclass that can be serialized using DataclassSerializer.

    Previously, Pydnatic has been used for defining objects as it provides built in support for serialization and
    deserialization. However, Pydantic object is slow compared to dataclass initialization, with tests showing 10x-100x
    slower performance. This is an issue if many objects are created, which can happen in during plan generation. Using
    the BaseModel.construct() is still not as fast as dataclass initiaization and it also makes for an awkward developer
    interface. Because of this, MF implements a simple custom serializer / deserializer to work with the built-in
    Python dataclass.

    The dataclass must have concrete types for all fields and not all types are supported. Please see implementation
    details in DataclassSerializer. Not adding post_init checks as there have been previous issues with slow object
    initialization.

    This is a concrete object as MyPy currently throws a type error if a Python dataclass is defined with an abstract
    parent class.
    """

    pass


SerializableDataclassT = TypeVar("SerializableDataclassT", bound=SerializableDataclass)


class DataclassSerializer:
    """Serializer that serializes SerializableDataclasses.

    Pydantic is useful for serialization, but it has issues serializing dataclasses. We've seen issues with forward
    references and recursion errors, but it's helpful for other data types. To serialize dataclasses, this class uses
    the type annotation defined in the dataclass to create a Pydantic model, and then uses Pydantic to serialize to a
    JSON string.
    """

    def __init__(self) -> None:  # noqa: D
        self._to_pydantic_type_converter = DataClassTypeToPydanticTypeConverter()

    def _convert_dataclass_instance_to_pydantic_model(
        self, object_type: Type, obj: Optional[AnyValueType] = None
    ) -> AnyValueType:
        if not _is_supported_field_type_in_serializable_dataclass(object_type):
            raise RuntimeError(f"Unsupported field type: {object_type}")
        elif _is_optional_type(object_type):
            if obj is None:
                return None
            optional_field_type_parameter = _get_type_parameter_for_optional(object_type)
            return self._convert_dataclass_instance_to_pydantic_model(
                object_type=optional_field_type_parameter, obj=obj
            )
        elif _is_sequence_like_tuple_type(object_type):
            if obj is None:
                return None

            tuple_field_type_parameter = _get_type_parameter_for_sequence_like_tuple_type(object_type)
            return tuple(
                self._convert_dataclass_instance_to_pydantic_model(
                    object_type=tuple_field_type_parameter,
                    obj=x,
                )
                for x in obj
            )
        elif issubclass(object_type, SerializableDataclass):
            if not dataclasses.is_dataclass(object_type):
                raise RuntimeError(f"{object_type} is not a dataclass")
            if not isinstance(obj, SerializableDataclass):
                raise RuntimeError(f"{obj} is not a SerializableDataclass")
            PydanticModel = self._to_pydantic_type_converter.to_pydantic_type(object_type)

            field_dict = _get_dataclass_field_definitions(object_type)
            field_values: Dict[str, AnyValueType] = {}
            for field_name, field_definition in field_dict.items():
                field_values[field_name] = self._convert_dataclass_instance_to_pydantic_model(
                    object_type=field_definition.annotated_field_type, obj=getattr(obj, field_name)
                )
            return PydanticModel(**field_values)

        return obj

    def pydantic_serialize(self, obj: SerializableDataclassT) -> str:  # noqa: D
        assert dataclasses.is_dataclass(obj)

        return self._convert_dataclass_instance_to_pydantic_model(
            # .__class__ seems to be the approach for new classes and there are differences with type(obj)
            object_type=obj.__class__,
            obj=obj,
        ).json()


class DataClassDeserializer:
    """Corresponding deserializer for datclasses that were serialized by DataClassSerializer."""

    def __init__(self) -> None:  # noqa: D
        self._to_pydantic_type_converter = DataClassTypeToPydanticTypeConverter()

    def _convert_field_in_pydantic_object_to_actual_object(
        self, field_type: Type, obj: Optional[AnyValueType] = None
    ) -> AnyValueType:
        if not _is_supported_field_type_in_serializable_dataclass(field_type):
            raise RuntimeError(f"Unsupported type: {field_type}")
        elif _is_optional_type(field_type):
            optional_field_type_parameter = _get_type_parameter_for_optional(field_type)
            if obj is None:
                return None

            return self._convert_field_in_pydantic_object_to_actual_object(
                field_type=optional_field_type_parameter,
                obj=obj,
            )
        elif _is_sequence_like_tuple_type(field_type):
            assert isinstance(obj, tuple)
            tuple_type_parameter = _get_type_parameter_for_sequence_like_tuple_type(field_type)
            return tuple(
                self._convert_field_in_pydantic_object_to_actual_object(
                    field_type=tuple_type_parameter,
                    obj=x,
                )
                for x in obj
            )
        elif issubclass(field_type, SerializableDataclass):
            logger.debug(f"Handling field_type={field_type} object={repr(obj)}")
            return self._construct_dataclass_from_pydantic_object(
                dataclass_type=field_type,
                obj=obj,
            )
        else:
            return obj

    def _construct_dataclass_from_pydantic_object(
        self, dataclass_type: Type[SerializableDataclassT], obj: BaseModel
    ) -> SerializableDataclassT:
        logger.debug(f"Constructing dataclass of type {dataclass_type} from {repr(obj)}")
        object_args = {}
        field_dict = _get_dataclass_field_definitions(dataclass_type)
        for field_name, field_definition in field_dict.items():
            object_args[field_name] = self._convert_field_in_pydantic_object_to_actual_object(
                field_type=field_definition.annotated_field_type,
                obj=getattr(obj, field_name),
            )

        return dataclass_type(**object_args)

    def pydantic_deserialize(  # noqa: D
        self, dataclass_type: Type[SerializableDataclassT], serialized_obj: str
    ) -> SerializableDataclassT:

        try:
            ClassAsPydantic = self._to_pydantic_type_converter.to_pydantic_type(dataclass_type)
            logger.debug(f"Serialized object for creation of {ClassAsPydantic} is {serialized_obj}")
            pydantic_object = ClassAsPydantic.parse_raw(serialized_obj)
            return self._construct_dataclass_from_pydantic_object(
                dataclass_type=dataclass_type,
                obj=pydantic_object,
            )

        except Exception as e:
            raise DataclassDeserializationError from e


class DataClassTypeToPydanticTypeConverter:  # noqa: D
    """Class that converts a SerializableDataclass into an equivalent Pydantic object.

    Includes caching to make it efficient.
    """

    def __init__(self) -> None:  # noqa: D
        self._dataclass_type_to_pydantic_type: Dict[Type, Type[BaseModel]] = {}

    def to_pydantic_type(self, dataclass_type: Type[SerializableDataclass]) -> Type[BaseModel]:  # noqa: D
        if dataclass_type not in self._dataclass_type_to_pydantic_type:
            self._dataclass_type_to_pydantic_type[
                dataclass_type
            ] = DataClassTypeToPydanticTypeConverter._convert_dataclass_type_to_pydantic_type(dataclass_type)
        return self._dataclass_type_to_pydantic_type[dataclass_type]

    @staticmethod
    def _convert_dataclass_type_to_pydantic_type(dataclass_type: Type) -> Type[BaseModel]:  # noqa: D
        logger.debug(f"Converting {dataclass_type.__name__} to a pydantic class")
        assert issubclass(dataclass_type, SerializableDataclass)
        assert dataclasses.is_dataclass(dataclass_type)

        field_dict = _get_dataclass_field_definitions(dataclass_type)

        # Maps the name of the field to (type of field, default value)
        fields_for_pydantic_model: Dict[str, Tuple[Type, AnyValueType]] = {}
        logger.debug(f"Need to add: {pformat_big_objects(field_dict.keys())}")
        for field_name, field_definition in field_dict.items():
            field_definition = DataClassTypeToPydanticTypeConverter._convert_nested_fields(field_definition)
            fields_for_pydantic_model[field_name] = field_definition.as_pydantic_field_tuple()
            logger.debug(f"Adding {field_name} with type {field_definition.annotated_field_type}")

        class_name = dataclass_type.__name__ + "AsPydantic"
        logger.debug(
            f"Creating Pydantic model {class_name} with fields:\n{pformat_big_objects(fields_for_pydantic_model)}"
        )
        pydantic_model = pydantic.create_model(class_name, **fields_for_pydantic_model)
        logger.debug(f"Finished creating Pydantic model {class_name}")
        logger.debug(f"Finished converting {dataclass_type.__name__} to a pydantic class")
        return pydantic_model

    @staticmethod
    def _convert_nested_fields(field_definition: FieldDefinition) -> FieldDefinition:
        """Recursively converts a given field definition into a fully serializable type specification

        The initial set of FieldDefinitions sourced from the dataclass might contain arbitrarily
        nested SerializableDataclass objects, which would need further parsing in order to be fully
        serializable via our Pydantic conversion. This method does that traversal and returns the complete
        set of Pydantic-compatible nested types.
        """
        if not _is_supported_field_type_in_serializable_dataclass(field_definition.annotated_field_type):
            raise RuntimeError(f"Unsupported type: {field_definition.annotated_field_type}")
        elif _is_optional_type(field_definition.annotated_field_type):
            optional_field_type_parameter = _get_type_parameter_for_optional(field_definition.annotated_field_type)
            converted_field_definition = DataClassTypeToPydanticTypeConverter._convert_nested_fields(
                FieldDefinition(field_type=optional_field_type_parameter)
            )
            return FieldDefinition(  # type: ignore[arg-type]
                Union[converted_field_definition.field_type, type(None)], field_definition.default_value
            )
        elif _is_sequence_like_tuple_type(field_definition.annotated_field_type):
            tuple_field_type_parameter = _get_type_parameter_for_sequence_like_tuple_type(
                field_definition.annotated_field_type
            )
            converted_field_definition = DataClassTypeToPydanticTypeConverter._convert_nested_fields(
                FieldDefinition(field_type=tuple_field_type_parameter)
            )
            return FieldDefinition(Tuple[converted_field_definition.field_type, ...], field_definition.default_value)
        elif issubclass(field_definition.annotated_field_type, SerializableDataclass):
            return FieldDefinition(
                field_type=DataClassTypeToPydanticTypeConverter._convert_dataclass_type_to_pydantic_type(
                    field_definition.annotated_field_type
                ),
                default_value=field_definition.default_value,
            )
        else:
            return field_definition


@dataclass(frozen=True)
class FieldDefinition:
    """Describes the field definition in a dataclass as described by the annotation.

    Note the default_value follows Pydantic semantics. A default value of ... indicates that there is no default, and
    as such a value must be provided for this field in the Pydantic BaseModel or Dataclass representation. We cannot
    use None for this, as None is a valid default for optional types.
    """

    field_type: AnyValueType
    default_value: Optional[AnyValueType] = ...

    def __post_init__(self) -> None:
        """Validate the combination of default_value and field_type"""
        if self.default_value is None:
            assert _is_optional_type(self.field_type), (
                f"Invalid default of None provided for field definition - type {self.field_type} will not allow it! "
                f"Use ... or dataclasses.MISSING instead"
            )

    def as_pydantic_field_tuple(self) -> Tuple[Type, AnyValueType]:
        """Pydantic fields can be initialized with a Tuple[Type, Any], where the second parameter is the default value"""
        default = self.default_value if self.default_value != dataclasses.MISSING else ...
        return (self.annotated_field_type, default)

    @property
    def annotated_field_type(self) -> Type:  # noqa: D
        return self.field_type



from metricflow.aggregation_properties import AggregationType
from metricflow.errors.errors import ModelTransformError
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule

ONE = "1"


class ConvertCountToSumRule(ModelTransformRule):
    """Converts any COUNT measures to SUM equivalent."""

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if measure.agg == AggregationType.COUNT:
                    if measure.expr is None:
                        raise ModelTransformError(
                            f"Measure '{measure.name}' uses a COUNT aggregation, which requires an expr to be provided. "
                            f"Provide 'expr: 1' if a count of all rows is desired."
                        )
                    if measure.expr != ONE:
                        # Just leave it as SUM(1) if we want to count all
                        measure.expr = f"CASE WHEN {measure.expr} IS NOT NULL THEN 1 ELSE 0 END"
                    measure.agg = AggregationType.SUM
        return model



from __future__ import annotations

from datetime import date
from typing import Union, Any

import pandas as pd

from metricflow.object_utils import assert_values_exhausted, ExtendedEnum


class TimeGranularity(ExtendedEnum):
    """For time dimensions, the smallest possible difference between two time values.

    Needed for calculating adjacency when merging 2 different time ranges.
    """

    # Names are used in parameters to DATE_TRUNC, so don't change them.
    # Values are used to convert user supplied strings to enums.
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"
    YEAR = "year"

    def to_int(self) -> int:
        """Convert to an int so that the size of the granularity can be easily compared."""
        if self is TimeGranularity.DAY:
            return 10
        elif self is TimeGranularity.WEEK:
            return 11
        elif self is TimeGranularity.MONTH:
            return 12
        elif self is TimeGranularity.QUARTER:
            return 13
        elif self is TimeGranularity.YEAR:
            return 14
        else:
            assert_values_exhausted(self)

    def is_smaller_than(self, other: "TimeGranularity") -> bool:  # noqa: D
        return self.to_int() < other.to_int()

    def is_smaller_than_or_equal(self, other: "TimeGranularity") -> bool:  # noqa: D
        return self.to_int() <= other.to_int()

    @property
    def offset_period(self) -> pd.offsets.DateOffset:
        """Offset object to use for adjusting by one granularity period."""
        # The type checker is throwing errors for some of those arguments, but they are valid.
        if self is TimeGranularity.DAY:
            return pd.offsets.DateOffset(days=1)  # type: ignore
        elif self is TimeGranularity.WEEK:
            return pd.offsets.DateOffset(weeks=1)  # type: ignore
        elif self is TimeGranularity.MONTH:
            return pd.offsets.DateOffset(months=1)
        elif self is TimeGranularity.QUARTER:
            return pd.offsets.DateOffset(months=3)
        elif self is TimeGranularity.YEAR:
            return pd.offsets.DateOffset(years=1)  # type: ignore
        else:
            assert_values_exhausted(self)

    @property
    def format_with_first_or_last(self) -> bool:
        """Indicates that this can only be calculated if query results display the first or last date of the period."""
        return self in [TimeGranularity.MONTH, TimeGranularity.QUARTER, TimeGranularity.YEAR]

    def is_period_start(self, date: Union[pd.Timestamp, date]) -> bool:  # noqa: D
        pd_date = pd.Timestamp(date)

        if self is TimeGranularity.DAY:
            return True
        elif self is TimeGranularity.WEEK:
            return ISOWeekDay.from_pandas_timestamp(pd_date).is_week_start
        elif self is TimeGranularity.MONTH:
            return pd_date.is_month_start
        elif self is TimeGranularity.QUARTER:
            return pd_date.is_quarter_start
        elif self is TimeGranularity.YEAR:
            return pd_date.is_year_start
        else:
            assert_values_exhausted(self)

    def is_period_end(self, date: Union[pd.Timestamp, date]) -> bool:  # noqa: D
        pd_date = pd.Timestamp(date)

        if self is TimeGranularity.DAY:
            return True
        elif self is TimeGranularity.WEEK:
            return ISOWeekDay.from_pandas_timestamp(pd_date).is_week_end
        elif self is TimeGranularity.MONTH:
            return pd_date.is_month_end
        elif self is TimeGranularity.QUARTER:
            return pd_date.is_quarter_end
        elif self is TimeGranularity.YEAR:
            return pd_date.is_year_end
        else:
            assert_values_exhausted(self)

    @property
    def period_begin_offset(  # noqa: D
        self,
    ) -> Union[pd.offsets.MonthBegin, pd.offsets.QuarterBegin, pd.offsets.Week, pd.offsets.YearBegin]:
        if self is TimeGranularity.DAY:
            raise ValueError(f"Can't get period start offset for TimeGranularity.{self.name}.")
        elif self is TimeGranularity.WEEK:
            return pd.offsets.Week(weekday=ISOWeekDay.MONDAY.pandas_value)
        elif self is TimeGranularity.MONTH:
            return pd.offsets.MonthBegin()
        elif self is TimeGranularity.QUARTER:
            return pd.offsets.QuarterBegin(startingMonth=1)
        elif self is TimeGranularity.YEAR:
            return pd.offsets.YearBegin()
        else:
            assert_values_exhausted(self)

    @property
    def period_end_offset(  # noqa: D
        self,
    ) -> Union[pd.offsets.MonthEnd, pd.offsets.QuarterEnd, pd.offsets.Week, pd.offsets.YearEnd]:
        if self is TimeGranularity.DAY:
            raise ValueError(f"Can't get period end offset for TimeGranularity.{self.name}.")
        elif self == TimeGranularity.WEEK:
            return pd.offsets.Week(weekday=ISOWeekDay.SUNDAY.pandas_value)
        elif self is TimeGranularity.MONTH:
            return pd.offsets.MonthEnd()
        elif self is TimeGranularity.QUARTER:
            return pd.offsets.QuarterEnd(startingMonth=3)
        elif self is TimeGranularity.YEAR:
            return pd.offsets.YearEnd()
        else:
            assert_values_exhausted(self)

    def adjust_to_start_of_period(self, date_to_adjust: pd.Timestamp, rollback: bool = True) -> pd.Timestamp:
        """Adjust to start of period if not at start already."""
        if rollback:
            return self.period_begin_offset.rollback(date_to_adjust)
        else:
            return self.period_begin_offset.rollforward(date_to_adjust)

    def adjust_to_end_of_period(self, date_to_adjust: pd.Timestamp, rollforward: bool = True) -> pd.Timestamp:
        """Adjust to end of period if not at end already."""
        if rollforward:
            return self.period_end_offset.rollforward(date_to_adjust)
        else:
            return self.period_end_offset.rollback(date_to_adjust)

    def match_start_or_end_of_period(self, date_to_match: pd.Timestamp, date_to_adjust: pd.Timestamp) -> pd.Timestamp:
        """Adjust date_to_adjust to be start or end of period based on if date_to_match is at start or end of period."""
        if self.is_period_start(date_to_match):
            return self.adjust_to_start_of_period(date_to_adjust)
        elif self.is_period_end(date_to_match):
            return self.adjust_to_end_of_period(date_to_adjust)
        else:
            raise ValueError(
                f"Expected `date_to_match` to fall at the start or end of the granularity period. Got '{date_to_match}' for granularity {self}."
            )

    def __lt__(self, other: Any) -> bool:  # type: ignore [misc] # noqa: D
        if not isinstance(other, TimeGranularity):
            return NotImplemented
        return self.to_int() < other.to_int()

    def __hash__(self) -> int:  # noqa: D
        return self.to_int()

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}.{self.name}"


class ISOWeekDay(ExtendedEnum):
    """Day of week values per ISO standard"""

    MONDAY = 1
    TUESDAY = 2
    WEDNESDAY = 3
    THURSDAY = 4
    FRIDAY = 5
    SATURDAY = 6
    SUNDAY = 7

    @staticmethod
    def from_pandas_timestamp(timestamp: pd.Timestamp) -> ISOWeekDay:
        """Factory for streamlining conversion from a Pandas Timestamp to an ISOWeekDay"""
        return ISOWeekDay(timestamp.isoweekday())

    @property
    def is_week_start(self) -> bool:
        """Return comparison of instance value against ISO standard start of week (Monday)"""
        return self is ISOWeekDay.MONDAY

    @property
    def is_week_end(self) -> bool:
        """Return comparison of instance value against ISO standard end of week (Sunday)"""
        return self is ISOWeekDay.SUNDAY

    @property
    def pandas_value(self) -> int:
        """Returns the pandas int value representation of the ISOWeekDay"""
        return self.value - 1


def string_to_time_granularity(s: str) -> TimeGranularity:  # noqa: D
    values = {item.value: item for item in TimeGranularity}
    return values[s]



"""File containing constants used by Configurations."""

CONFIG_PATH_KEY = "MF_CONFIG_DIR"
CONFIG_DWH_DIALECT = "dwh_dialect"
CONFIG_DWH_HOST = "dwh_host"
CONFIG_DWH_PORT = "dwh_port"
CONFIG_DWH_USER = "dwh_user"
CONFIG_DWH_PASSWORD = "dwh_password"
CONFIG_DWH_DB = "dwh_database"
CONFIG_DWH_SCHEMA = "dwh_schema"
CONFIG_DWH_CREDS_PATH = "dwh_path_to_creds"
CONFIG_DWH_WAREHOUSE = "dwh_warehouse"
CONFIG_DWH_PROJECT_ID = "dwh_project_id"
CONFIG_EMAIL = "email"
CONFIG_MODEL_PATH = "model_path"
CONFIG_DWH_HTTP_PATH = "dwh_http_path"
CONFIG_DWH_ACCESS_TOKEN = "dwh_access_token"
CONFIG_DBT_REPO = "dbt_repo"
CONFIG_DBT_PROFILE = "dbt_profile"
CONFIG_DBT_TARGET = "dbt_target"



"""Various classes that perform a transformation on instance sets."""

from __future__ import annotations

import logging
from collections import OrderedDict
from itertools import chain
from typing import List, Sequence, Optional, Dict, Tuple

from metricflow.aggregation_properties import AggregationState
from metricflow.instances import (
    MdoInstance,
    DimensionInstance,
    IdentifierInstance,
    MetricInstance,
    MeasureInstance,
    InstanceSet,
    InstanceSetTransform,
    TimeDimensionInstance,
)
from metricflow.protocols.semantics import DataSourceSemanticsAccessor
from metricflow.object_utils import assert_exactly_one_arg_set
from metricflow.plan_conversion.select_column_gen import SelectColumnSet
from metricflow.specs import (
    MeasureSpec,
    MetricInputMeasureSpec,
    InstanceSpec,
    IdentifierSpec,
    DimensionSpec,
    ColumnAssociationResolver,
    TimeDimensionSpec,
    LinklessIdentifierSpec,
    LinkableInstanceSpec,
    IdentifierReference,
)
from metricflow.sql.sql_exprs import (
    SqlColumnReferenceExpression,
    SqlColumnReference,
    SqlFunctionExpression,
)
from metricflow.sql.sql_plan import SqlSelectColumn

logger = logging.getLogger(__name__)


class CreateSelectColumnsForInstances(InstanceSetTransform[SelectColumnSet]):
    """Create select column expressions that will express all instances in the set.

    It assumes that the column names of the instances are represented by the supplied column association resolver and
    come from the given table alias.
    """

    def __init__(
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        output_to_input_column_mapping: Optional[OrderedDict[str, str]] = None,
    ) -> None:
        """Initializer.

        Args:
            table_alias: the table alias to select columns from
            column_resolver: resolver to name columns.
            output_to_input_column_mapping: if specified, use these columns in the input for the given output columns.
        """
        self._table_alias = table_alias
        self._column_resolver = column_resolver
        self._output_to_input_column_mapping = output_to_input_column_mapping or OrderedDict()

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D
        metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metric_instances])
        )
        measure_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.measure_instances])
        )
        dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.dimension_instances])
        )
        time_dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.time_dimension_instances])
        )
        identifier_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.identifier_instances])
        )
        return SelectColumnSet(
            metric_columns=metric_cols,
            measure_columns=measure_cols,
            dimension_columns=dimension_cols,
            time_dimension_columns=time_dimension_cols,
            identifier_columns=identifier_cols,
        )

    def _make_sql_column_expression(
        self,
        element_instance: MdoInstance,
    ) -> List[SqlSelectColumn]:
        """Convert one element instance into a SQL column"""

        # Do a sanity check to make sure that there's a 1:1 mapping between the columns associations generated by the
        # column resolver based on the spec, and the columns that are already associated with the instance.
        expected_column_associations = element_instance.spec.column_associations(self._column_resolver)
        existing_column_associations = element_instance.associated_columns

        # Dict between the expected column name and the corresponding column in the existing columns
        column_matches: Dict[str, List[str]] = {
            expected_column.column_name: [
                col.column_name
                for col in existing_column_associations
                if col.column_correlation_key == expected_column.column_correlation_key
            ]
            for expected_column in expected_column_associations
        }

        # Assert a 1:1 mapping between expected and existing
        assert all([len(x) == 1 for x in column_matches.values()]), (
            f"Did not find exactly one match for each expected column associations.  "
            f"Expected -> existing mappings: {column_matches}"
        )
        existing_names = set([col.column_name for col in existing_column_associations])
        mapped_names = set()
        mapped_cols: List[str] = []
        for mapped_cols in column_matches.values():
            mapped_names.update([col_name for col_name in mapped_cols])
        assert existing_names == mapped_names, (
            f"Not all existing columns were mapped. Existing: {existing_names}.  Mapped: {mapped_cols}, "
            f"{expected_column_associations} -- {existing_column_associations}"
        )

        select_columns = []
        for expected_name, mapped_cols in column_matches.items():
            input_column_name = mapped_cols[0]
            output_column_name = expected_name

            if output_column_name in self._output_to_input_column_mapping:
                input_column_name = self._output_to_input_column_mapping[output_column_name]
            select_columns.append(
                SqlSelectColumn(
                    expr=SqlColumnReferenceExpression(SqlColumnReference(self._table_alias, input_column_name)),
                    column_alias=output_column_name,
                )
            )
        return select_columns


class CreateSelectColumnsWithMeasuresAggregated(CreateSelectColumnsForInstances):
    """Create select columns of the form "fct_bookings.bookings AS bookings" for all the instances.

    However, for measure columns, convert them into expressions like "SUM(fct_bookings.bookings) AS bookings" so that
    the resulting expressions can be used for aggregations.

    Also add an output alias that conforms to the alias
    """

    def __init__(  # noqa: D
        self,
        table_alias: str,
        column_resolver: ColumnAssociationResolver,
        data_source_semantics: DataSourceSemanticsAccessor,
        metric_input_measure_specs: Sequence[MetricInputMeasureSpec],
    ) -> None:
        self._data_source_semantics = data_source_semantics
        self.metric_input_measure_specs = metric_input_measure_specs
        super().__init__(table_alias=table_alias, column_resolver=column_resolver)

    def _make_sql_column_expression_to_aggregate_measures(
        self, measure_instances: Tuple[MeasureInstance, ...]
    ) -> List[SqlSelectColumn]:
        output_columns: List[SqlSelectColumn] = []
        aliased_input_specs = [spec for spec in self.metric_input_measure_specs if spec.alias]
        for instance in measure_instances:
            matches = [spec for spec in aliased_input_specs if spec.measure_spec == instance.spec]
            if matches:
                aliased_spec = matches[0]
                aliased_input_specs.remove(aliased_spec)
                output_measure_spec = aliased_spec.post_aggregation_spec
            else:
                output_measure_spec = instance.spec

            output_columns.append(
                self._make_sql_column_expression_to_aggregate_measure(
                    measure_instance=instance, output_measure_spec=output_measure_spec
                )
            )

        return output_columns

    def _make_sql_column_expression_to_aggregate_measure(  # noqa: D
        self, measure_instance: MeasureInstance, output_measure_spec: MeasureSpec
    ) -> SqlSelectColumn:
        """Convert one measure instance into a SQL column"""
        # Get the column name of the measure in the table that we're reading from
        column_name_in_table = measure_instance.associated_column.column_name

        # Create an expression that will aggregate the given measure.
        # Figure out the aggregation function for the measure.
        measure = self._data_source_semantics.get_measure(measure_instance.spec.as_reference)
        aggregation_type = measure.agg

        expression_to_get_measure = SqlColumnReferenceExpression(
            SqlColumnReference(self._table_alias, column_name_in_table)
        )

        expression_to_aggregate_measure = SqlFunctionExpression.from_aggregation_type(
            aggregation_type=aggregation_type, sql_column_expression=expression_to_get_measure
        )

        # Get the output column name from the measure/alias

        new_column_association_for_aggregated_measure = self._column_resolver.resolve_measure_spec(output_measure_spec)
        new_column_name_for_aggregated_measure = new_column_association_for_aggregated_measure.column_name

        return SqlSelectColumn(
            expr=expression_to_aggregate_measure,
            column_alias=new_column_name_for_aggregated_measure,
        )

    def transform(self, instance_set: InstanceSet) -> SelectColumnSet:  # noqa: D
        metric_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.metric_instances])
        )

        measure_cols = self._make_sql_column_expression_to_aggregate_measures(instance_set.measure_instances)
        dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.dimension_instances])
        )
        time_dimension_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.time_dimension_instances])
        )
        identifier_cols = list(
            chain.from_iterable([self._make_sql_column_expression(x) for x in instance_set.identifier_instances])
        )
        return SelectColumnSet(
            metric_columns=metric_cols,
            measure_columns=measure_cols,
            dimension_columns=dimension_cols,
            time_dimension_columns=time_dimension_cols,
            identifier_columns=identifier_cols,
        )


class AddLinkToLinkableElements(InstanceSetTransform[InstanceSet]):
    """Return a new instance set where the all linkable elements in the set have a new link added.

    e.g. "country" -> "user_id__country" after a data set has been joined by identifier.
    """

    def __init__(self, join_on_identifier: LinklessIdentifierSpec) -> None:  # noqa: D
        self._join_on_identifier = join_on_identifier

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        assert len(instance_set.metric_instances) == 0, "Can't add links to instance sets with metrics"
        assert len(instance_set.measure_instances) == 0, "Can't add links to instance sets with measures"

        # Handle dimension instances
        dimension_instances_with_additional_link = []
        for dimension_instance in instance_set.dimension_instances:
            # The new dimension spec should include the join on identifier.
            transformed_dimension_spec_from_right = DimensionSpec(
                element_name=dimension_instance.spec.element_name,
                identifier_links=self._join_on_identifier.as_linkless_prefix + dimension_instance.spec.identifier_links,
            )
            dimension_instances_with_additional_link.append(
                DimensionInstance(
                    associated_columns=dimension_instance.associated_columns,
                    defined_from=dimension_instance.defined_from,
                    spec=transformed_dimension_spec_from_right,
                )
            )

        # Handle time dimension instances
        time_dimension_instances_with_additional_link = []
        for time_dimension_instance in instance_set.time_dimension_instances:
            # The new dimension spec should include the join on identifier.
            transformed_time_dimension_spec_from_right = TimeDimensionSpec(
                element_name=time_dimension_instance.spec.element_name,
                identifier_links=(
                    (IdentifierReference(element_name=self._join_on_identifier.element_name),)
                    + time_dimension_instance.spec.identifier_links
                ),
                time_granularity=time_dimension_instance.spec.time_granularity,
            )
            time_dimension_instances_with_additional_link.append(
                TimeDimensionInstance(
                    associated_columns=time_dimension_instance.associated_columns,
                    defined_from=time_dimension_instance.defined_from,
                    spec=transformed_time_dimension_spec_from_right,
                )
            )

        # Handle identifier instances
        identifier_instances_with_additional_link = []
        for identifier_instance in instance_set.identifier_instances:
            # Don't include adding the identifier link to the same identifier.
            # Otherwise, you would create "user_id__user_id", which is confusing.
            if identifier_instance.spec == self._join_on_identifier:
                continue
            # The new identifier spec should include the join on identifier.
            transformed_identifier_spec_from_right = IdentifierSpec(
                element_name=identifier_instance.spec.element_name,
                identifier_links=self._join_on_identifier.as_linkless_prefix
                + identifier_instance.spec.identifier_links,
            )
            identifier_instances_with_additional_link.append(
                IdentifierInstance(
                    associated_columns=identifier_instance.associated_columns,
                    defined_from=identifier_instance.defined_from,
                    spec=transformed_identifier_spec_from_right,
                )
            )

        return InstanceSet(
            measure_instances=(),
            dimension_instances=tuple(dimension_instances_with_additional_link),
            time_dimension_instances=tuple(time_dimension_instances_with_additional_link),
            identifier_instances=tuple(identifier_instances_with_additional_link),
            metric_instances=(),
        )


class FilterLinkableInstancesWithLeadingLink(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that have a specified leading link removed.

    e.g. Remove "listing__country" if the specified link is "listing".
    """

    def __init__(  # noqa: D
        self,
        identifier_link: LinklessIdentifierSpec,
    ) -> None:
        """Constructor.

        Args:
            identifier_link: Remove elements with this link as the first element in "identifier_links"
        """
        self._identifier_link = identifier_link

    def _should_pass(self, linkable_spec: LinkableInstanceSpec) -> bool:  # noqa: D
        return (
            len(linkable_spec.identifier_links) == 0
            or LinklessIdentifierSpec.from_reference(linkable_spec.identifier_links[0]) != self._identifier_link
        )

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D

        # Normal to not filter anything if the instance set has no instances with links.
        filtered_dimension_instances = tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec))
        filtered_time_dimension_instances = tuple(
            x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
        )
        filtered_identifier_instances = tuple(x for x in instance_set.identifier_instances if self._should_pass(x.spec))

        output = InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=filtered_dimension_instances,
            time_dimension_instances=filtered_time_dimension_instances,
            identifier_instances=filtered_identifier_instances,
            metric_instances=instance_set.metric_instances,
        )
        return output


class FilterElements(InstanceSetTransform[InstanceSet]):
    """Return an instance set with the elements that don't match any of the pass specs removed."""

    def __init__(  # noqa: D
        self,
        include_specs: Optional[Sequence[InstanceSpec]] = None,
        exclude_specs: Optional[Sequence[InstanceSpec]] = None,
    ) -> None:
        """Constructor.

        Args:
            include_specs: If specified, pass only instances matching these specs.
            exclude_specs: If specified, pass only instances not matching these specs.
        """
        assert_exactly_one_arg_set(include_specs=include_specs, exclude_specs=exclude_specs)
        self._include_specs = include_specs
        self._exclude_specs = exclude_specs

    def _should_pass(self, element_spec: InstanceSpec) -> bool:  # noqa: D
        # TODO: Use better matching function
        if self._include_specs:
            return any(x == element_spec for x in self._include_specs)
        elif self._exclude_specs:
            return not any(x == element_spec for x in self._exclude_specs)
        assert False

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        # Sanity check to make sure the specs are in the instance set

        if self._include_specs:
            for include_spec in self._include_specs:
                if include_spec not in instance_set.spec_set.all_specs:
                    raise RuntimeError(
                        f"Include spec {include_spec} is not in the spec set {instance_set.spec_set} - "
                        f"check if this node was constructed correctly."
                    )
        elif self._exclude_specs:
            for exclude_spec in self._exclude_specs:
                if exclude_spec not in instance_set.spec_set.all_specs:
                    raise RuntimeError(
                        f"Exclude spec {exclude_spec} is not in the spec set {instance_set.spec_set} - "
                        f"check if this node was constructed correctly."
                    )
        else:
            assert False, "Include specs or exclude specs should have been specified."

        output = InstanceSet(
            measure_instances=tuple(x for x in instance_set.measure_instances if self._should_pass(x.spec)),
            dimension_instances=tuple(x for x in instance_set.dimension_instances if self._should_pass(x.spec)),
            time_dimension_instances=tuple(
                x for x in instance_set.time_dimension_instances if self._should_pass(x.spec)
            ),
            identifier_instances=tuple(x for x in instance_set.identifier_instances if self._should_pass(x.spec)),
            metric_instances=tuple(x for x in instance_set.metric_instances if self._should_pass(x.spec)),
        )
        return output


class ChangeMeasureAggregationState(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all measures are set as a different aggregation state."""

    def __init__(self, aggregation_state_changes: Dict[AggregationState, AggregationState]) -> None:
        """Constructor.

        Args:
            aggregation_state_changes: key is old aggregation state, value is the new aggregation state.
        """
        self._aggregation_state_changes = aggregation_state_changes

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        for measure_instance in instance_set.measure_instances:
            assert measure_instance.aggregation_state in self._aggregation_state_changes, (
                f"Aggregation state: {measure_instance.aggregation_state} not handled in change dict: "
                f"{self._aggregation_state_changes}"
            )

        # Copy the measures, but just change the aggregation state to COMPLETE.
        measure_instances = tuple(
            MeasureInstance(
                associated_columns=x.associated_columns,
                defined_from=x.defined_from,
                aggregation_state=self._aggregation_state_changes[x.aggregation_state],
                spec=x.spec,
            )
            for x in instance_set.measure_instances
        )
        return InstanceSet(
            measure_instances=measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            identifier_instances=instance_set.identifier_instances,
            metric_instances=instance_set.metric_instances,
        )


class AliasAggregatedMeasures(InstanceSetTransform[InstanceSet]):
    """Returns a new instance set where all measures have been assigned an alias spec"""

    def __init__(self, metric_input_measure_specs: Sequence[MetricInputMeasureSpec]):
        """Initializer stores the input specs, which contain the aliases for each measure

        Note this class only works if used in conjunction with an AggregateMeasuresNode that has been generated
        by querying a single data source for a single set of aggregated measures. This is currently enforced
        by the structure of the DataflowPlanBuilder, which ensures each AggregateMeasuresNode corresponds to
        a single data source set of measures for a single metric, and that these outputs will then be
        combinded via joins.
        """
        self.metric_input_measure_specs = metric_input_measure_specs

    def _alias_measure_instances(self, measure_instances: Tuple[MeasureInstance, ...]) -> Tuple[MeasureInstance, ...]:
        """Update all measure instances with aliases, if any are found in the input spec set"""
        aliased_instances: List[MeasureInstance] = []
        aliased_input_specs = [spec for spec in self.metric_input_measure_specs if spec.alias]
        for instance in measure_instances:
            matches = [spec for spec in aliased_input_specs if spec.measure_spec == instance.spec]
            assert (
                len(matches) < 2
            ), f"Found duplicate aliased measure spec matches: {matches} for measure instance {instance}. "
            "We should always have 0 or 1 matches, or else we might pass the wrong aggregated measures to the "
            "downstream metric computation expression!"
            if matches:
                aliased_spec = matches[0]
                aliased_input_specs.remove(aliased_spec)
                measure_spec = aliased_spec.post_aggregation_spec
            else:
                measure_spec = instance.spec

            aliased_instances.append(
                MeasureInstance(
                    associated_columns=instance.associated_columns,
                    spec=measure_spec,
                    aggregation_state=instance.aggregation_state,
                    defined_from=instance.defined_from,
                )
            )

        return tuple(aliased_instances)

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        return InstanceSet(
            measure_instances=self._alias_measure_instances(instance_set.measure_instances),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            identifier_instances=instance_set.identifier_instances,
            metric_instances=instance_set.metric_instances,
        )


class AddMetrics(InstanceSetTransform[InstanceSet]):
    """Adds the given metric instances to the instance set"""

    def __init__(self, metric_instances: List[MetricInstance]) -> None:  # noqa: D
        self._metric_instances = metric_instances

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        return InstanceSet(
            measure_instances=instance_set.measure_instances,
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            identifier_instances=instance_set.identifier_instances,
            metric_instances=instance_set.metric_instances + tuple(self._metric_instances),
        )


class RemoveMeasures(InstanceSetTransform[InstanceSet]):
    """Remove measures from the instance set."""

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        return InstanceSet(
            measure_instances=(),
            dimension_instances=instance_set.dimension_instances,
            time_dimension_instances=instance_set.time_dimension_instances,
            identifier_instances=instance_set.identifier_instances,
            metric_instances=instance_set.metric_instances,
        )


class ChangeAssociatedColumns(InstanceSetTransform[InstanceSet]):
    """Change the columns associated with instances to the one specified by the resolver."""

    def __init__(self, column_association_resolver: ColumnAssociationResolver) -> None:  # noqa: D
        self._column_association_resolver = column_association_resolver

    def transform(self, instance_set: InstanceSet) -> InstanceSet:  # noqa: D
        output_measure_instances = []
        for input_measure_instance in instance_set.measure_instances:
            output_measure_instances.append(
                MeasureInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_measure_spec(
                            measure_spec=input_measure_instance.spec
                        ),
                    ),
                    spec=input_measure_instance.spec,
                    defined_from=input_measure_instance.defined_from,
                    aggregation_state=input_measure_instance.aggregation_state,
                )
            )

        output_dimension_instances = []
        for input_dimension_instance in instance_set.dimension_instances:
            output_dimension_instances.append(
                DimensionInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_dimension_spec(
                            dimension_spec=input_dimension_instance.spec
                        ),
                    ),
                    spec=input_dimension_instance.spec,
                    defined_from=input_dimension_instance.defined_from,
                )
            )

        output_time_dimension_instances = []
        for input_time_dimension_instance in instance_set.time_dimension_instances:
            output_time_dimension_instances.append(
                TimeDimensionInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_time_dimension_spec(
                            time_dimension_spec=input_time_dimension_instance.spec
                        ),
                    ),
                    spec=input_time_dimension_instance.spec,
                    defined_from=input_time_dimension_instance.defined_from,
                )
            )

        output_identifier_instances = []
        for input_identifier_instance in instance_set.identifier_instances:
            output_identifier_instances.append(
                IdentifierInstance(
                    associated_columns=self._column_association_resolver.resolve_identifier_spec(
                        identifier_spec=input_identifier_instance.spec
                    ),
                    spec=input_identifier_instance.spec,
                    defined_from=input_identifier_instance.defined_from,
                )
            )

        output_metric_instances = []
        for input_metric_instance in instance_set.metric_instances:
            output_metric_instances.append(
                MetricInstance(
                    associated_columns=(
                        self._column_association_resolver.resolve_metric_spec(metric_spec=input_metric_instance.spec),
                    ),
                    spec=input_metric_instance.spec,
                    defined_from=input_metric_instance.defined_from,
                )
            )

        return InstanceSet(
            measure_instances=tuple(output_measure_instances),
            dimension_instances=tuple(output_dimension_instances),
            time_dimension_instances=tuple(output_time_dimension_instances),
            identifier_instances=tuple(output_identifier_instances),
            metric_instances=tuple(output_metric_instances),
        )


def create_select_columns_for_instance_sets(
    column_resolver: ColumnAssociationResolver,
    table_alias_to_instance_set: OrderedDict[str, InstanceSet],
) -> Tuple[SqlSelectColumn, ...]:
    """Creates select columns for instance sets coming from multiple table as defined in table_alias_to_instance_set.

    Used in cases where you join multiple tables and need to render select columns to access all of those.
    """
    column_set = SelectColumnSet()
    for table_alias, instance_set in table_alias_to_instance_set.items():
        column_set = column_set.merge(
            instance_set.transform(
                CreateSelectColumnsForInstances(
                    table_alias=table_alias,
                    column_resolver=column_resolver,
                )
            )
        )

    return column_set.as_tuple()



import logging
from typing import Optional

from metricflow.model.objects.data_source import DataSource
from metricflow.model.objects.elements.dimension import DimensionType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule
from metricflow.references import TimeDimensionReference

logger = logging.getLogger(__name__)


class SetMeasureAggregationTimeDimensionRule(ModelTransformRule):
    """Sets the aggregation time dimension for measures to the primary time dimension if not defined."""

    @staticmethod
    def _find_primary_time_dimension(data_source: DataSource) -> Optional[TimeDimensionReference]:
        for dimension in data_source.dimensions:
            if dimension.type == DimensionType.TIME and dimension.type_params and dimension.type_params.is_primary:
                return dimension.time_dimension_reference
        return None

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        for data_source in model.data_sources:
            primary_time_dimension_reference = SetMeasureAggregationTimeDimensionRule._find_primary_time_dimension(
                data_source
            )

            if not primary_time_dimension_reference:
                # Dimension data sources won't have a primary time dimension.
                continue

            for measure in data_source.measures:
                if not measure.agg_time_dimension:
                    measure.agg_time_dimension = primary_time_dimension_reference.element_name

        return model



"""Nodes used in defining SQL expressions."""

from __future__ import annotations

import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import List, Generic, Sequence, Optional, Tuple, Dict

from metricflow.aggregation_properties import AggregationType
from metricflow.dag.mf_dag import DagNode, DisplayedProperty, NodeId
from metricflow.dag.id_generation import (
    SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX,
    SQL_EXPR_FUNCTION_ID_PREFIX,
    SQL_EXPR_STRING_ID_PREFIX,
    SQL_EXPR_COMPARISON_ID_PREFIX,
    SQL_EXPR_NULL_PREFIX,
    SQL_EXPR_LOGICAL_OPERATOR_PREFIX,
    SQL_EXPR_STRING_LITERAL_PREFIX,
    SQL_EXPR_IS_NULL_PREFIX,
    SQL_EXPR_DATE_TRUNC,
    SQL_EXPR_RATIO_COMPUTATION,
    SQL_EXPR_BETWEEN_PREFIX,
)
from metricflow.sql.sql_bind_parameters import SqlBindParameters
from metricflow.visitor import Visitable, VisitorOutputT
from metricflow.time.time_granularity import TimeGranularity
from metricflow.object_utils import assert_values_exhausted, flatten_nested_sequence


class SqlExpressionNode(DagNode, Generic[VisitorOutputT], Visitable, ABC):
    """An SQL expression like my_table.my_column, CONCAT(a, b) or 1 + 1 that evaluates to a value."""

    def __init__(self, node_id: NodeId, parent_nodes: List[SqlExpressionNode]) -> None:  # noqa: D
        self._parent_nodes = parent_nodes
        super().__init__(node_id=node_id)

    @property
    @abstractmethod
    def requires_parenthesis(self) -> bool:
        """Should expression needs be rendered with parenthesis when rendering inside other expressions.

        Useful for string expressions where we can't infer the structure. For example, in rendering

        SqlMathExpression(operator="*", left_expr=SqlStringExpression("a"), right_expr=SqlStringExpression("b + c")

        this can be used to differentiate between

        a * b + c vs. a * (b + c)
        """
        pass

    @abstractmethod
    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:
        """Called when a visitor needs to visit this node."""
        pass

    @property
    def execution_parameters(self) -> SqlBindParameters:
        """Execution parameters when running a query containing this expression.

        * See: https://docs.sqlalchemy.org/en/14/core/tutorial.html#using-textual-sql
        * Generally only defined for string expressions.
        """
        return SqlBindParameters()

    @property
    def parent_nodes(self) -> Sequence[SqlExpressionNode]:  # noqa: D
        return self._parent_nodes

    @property
    def as_column_reference_expression(self) -> Optional[SqlColumnReferenceExpression]:
        """If this is a column reference expression, return self."""
        return None

    @property
    def as_string_expression(self) -> Optional[SqlStringExpression]:
        """If this is a string expression, return self."""
        return None

    @abstractmethod
    def rewrite(
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        """Return the same semantic expression but with re-written according to the input.

        Args:
            column_replacements: Replaces column references according to this map.
            should_render_table_alias: Change if table aliases should be rendered for column reference expressions.
        """
        pass

    @property
    @abstractmethod
    def lineage(self) -> SqlExpressionTreeLineage:
        """Returns all nodes in the paths from this node to the root nodes"""
        pass

    def _parents_match(self, other: SqlExpressionNode) -> bool:  # noqa: D
        return all(x == y for x, y in itertools.zip_longest(self.parent_nodes, other.parent_nodes))

    @abstractmethod
    def matches(self, other: SqlExpressionNode) -> bool:
        """Similar to equals - returns true if these expressions are equivalent."""
        pass


@dataclass(frozen=True)
class SqlExpressionTreeLineage:
    """Captures the lineage of an expression node - contains itself and all ancestor nodes."""

    string_exprs: Tuple[SqlStringExpression, ...] = ()
    function_exprs: Tuple[SqlFunctionExpression, ...] = ()
    column_reference_exprs: Tuple[SqlColumnReferenceExpression, ...] = ()
    column_alias_reference_exprs: Tuple[SqlColumnAliasReferenceExpression, ...] = ()
    other_exprs: Tuple[SqlExpressionNode, ...] = ()

    @staticmethod
    def combine(lineages: Sequence[SqlExpressionTreeLineage]) -> SqlExpressionTreeLineage:
        """Combine multiple lineages into one lineage, without de-duping."""
        return SqlExpressionTreeLineage(
            string_exprs=flatten_nested_sequence(tuple(x.string_exprs for x in lineages)),
            function_exprs=flatten_nested_sequence(tuple(x.function_exprs for x in lineages)),
            column_reference_exprs=flatten_nested_sequence(tuple(x.column_reference_exprs for x in lineages)),
            column_alias_reference_exprs=flatten_nested_sequence(
                tuple(x.column_alias_reference_exprs for x in lineages)
            ),
            other_exprs=flatten_nested_sequence(tuple(x.other_exprs for x in lineages)),
        )

    @property
    def contains_string_exprs(self) -> bool:  # noqa: D
        return len(self.string_exprs) > 0

    @property
    def contains_column_alias_exprs(self) -> bool:  # noqa: D
        return len(self.column_alias_reference_exprs) > 0

    @property
    def contains_ambiguous_exprs(self) -> bool:  # noqa: D
        return self.contains_string_exprs or self.contains_column_alias_exprs


class SqlColumnReplacements:
    """When re-writing column references in expressions, this storing the mapping."""

    def __init__(self, column_replacements: Dict[SqlColumnReference, SqlExpressionNode]) -> None:  # noqa: D
        self._column_replacements = column_replacements

    def get_replacement(self, column_reference: SqlColumnReference) -> Optional[SqlExpressionNode]:  # noqa: D
        return self._column_replacements.get(column_reference)


class SqlExpressionNodeVisitor(Generic[VisitorOutputT], ABC):
    """A visitor to help visit the nodes of an expression.

    See similar visitor DataflowPlanVisitor.
    """

    @abstractmethod
    def visit_string_expr(self, node: SqlStringExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_column_reference_expr(self, node: SqlColumnReferenceExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_column_alias_reference_expr(self, node: SqlColumnAliasReferenceExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_comparison_expr(self, node: SqlComparisonExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_function_expr(self, node: SqlFunctionExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_null_expr(self, node: SqlNullExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_logical_expr(self, node: SqlLogicalExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_string_literal_expr(self, node: SqlStringLiteralExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_is_null_expr(self, node: SqlIsNullExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_cast_to_timestamp_expr(self, node: SqlCastToTimestampExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_date_trunc_expr(self, node: SqlDateTruncExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_time_delta_expr(self, node: SqlTimeDeltaExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_ratio_computation_expr(self, node: SqlRatioComputationExpression) -> VisitorOutputT:  # noqa: D
        pass

    @abstractmethod
    def visit_between_expr(self, node: SqlBetweenExpression) -> VisitorOutputT:  # noqa: D
        pass


class SqlStringExpression(SqlExpressionNode):
    """An SQL expression in a string format, so it lacks information about the structure.

    These are convenient to use, but because structure is lacking, it can't be easily handled for DB rendering and can
    impede optimizations.
    """

    def __init__(
        self,
        sql_expr: str,
        execution_parameters: Optional[SqlBindParameters] = None,
        requires_parenthesis: bool = True,
        used_columns: Optional[Tuple[str, ...]] = None,
    ) -> None:
        """Constructor.

        Args:
            sql_expr: The SQL in string form.
            execution_parameters: See SqlExpressionNode.execution_parameters
            requires_parenthesis: Whether this should be rendered with () if nested in another expression.
            used_columns: If set, indicates that the expression represented by the string only uses those columns. e.g.
            sql_expr="a + b", used_columns=["a", "b"]. This may be used by optimizers, and if specified, it must be
            complete. e.g. sql_expr="a + b + c", used_columns=["a", "b"] will cause problems.
        """
        self._sql_expr = sql_expr
        self._execution_parameters = execution_parameters or SqlBindParameters()
        self._requires_parenthesis = requires_parenthesis
        self._used_columns = used_columns
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_STRING_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_string_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return f"String SQL Expression: {self._sql_expr}"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [DisplayedProperty("sql_expr", self._sql_expr)]

    @property
    def sql_expr(self) -> str:  # noqa: D
        return self._sql_expr

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return self._requires_parenthesis

    @property
    def execution_parameters(self) -> SqlBindParameters:  # noqa: D
        return self._execution_parameters

    @property
    def used_columns(self) -> Optional[Tuple[str, ...]]:  # noqa: D
        return self._used_columns

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(node_id={self.node_id} sql_expr={self.sql_expr})"

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        if column_replacements:
            raise NotImplementedError()
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage(string_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlStringExpression):
            return False
        return (
            self.sql_expr == other.sql_expr
            and self.used_columns == other.used_columns
            and self.execution_parameters == other.execution_parameters
        )

    def as_string_expression(self) -> Optional[SqlStringExpression]:
        """If this is a string expression, return self."""
        return self


class SqlStringLiteralExpression(SqlExpressionNode):
    """A string literal like 'foo'. It shouldn't include delimiters as it should be added during rendering."""

    def __init__(self, literal_value: str) -> None:  # noqa: D
        self._literal_value = literal_value
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_STRING_LITERAL_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_string_literal_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return f"String Literal: {self._literal_value}"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [DisplayedProperty("value", self._literal_value)]

    @property
    def literal_value(self) -> str:  # noqa: D
        return self._literal_value

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    @property
    def execution_parameters(self) -> SqlBindParameters:  # noqa: D
        return SqlBindParameters()

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(node_id={self.node_id}, literal_value={self.literal_value})"

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlStringLiteralExpression):
            return False
        return self.literal_value == other.literal_value


@dataclass(frozen=True)
class SqlColumnReference:
    """Used with string expressions to specify what columns are referred to in the string expression."""

    table_alias: str
    column_name: str


class SqlColumnReferenceExpression(SqlExpressionNode):
    """An expression that evaluates to the value of a column in one of the sources in the select query.

    e.g. my_table.my_column
    """

    def __init__(self, col_ref: SqlColumnReference, should_render_table_alias: bool = True) -> None:
        """Constructor.

        Args:
            col_ref: the associated column reference.
            should_render_table_alias: When converting this to SQL text, whether the table alias needed to be included.
            e.g. "foo.bar" vs "bar".
        """
        self._col_ref = col_ref
        self._should_render_table_alias = should_render_table_alias
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_column_reference_expr(self)

    @property
    def col_ref(self) -> SqlColumnReference:  # noqa: D
        return self._col_ref

    @property
    def description(self) -> str:  # noqa: D
        return f"Column: {self.col_ref}"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [DisplayedProperty("col_ref", self.col_ref)]

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    @property
    def as_column_reference_expression(self) -> Optional[SqlColumnReferenceExpression]:  # noqa:
        return self

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:

        # TODO: Hack to work around the fact our test data set contains "user", which is a reserved keyword.
        # We should migrate "user" -> "user_id" in the test set.
        # This will force "user" to be rendered as "table_alias.user"
        if self.col_ref.column_name == "user":
            should_render_table_alias = True

        if column_replacements:
            replacement = column_replacements.get_replacement(self.col_ref)
            if replacement:
                if should_render_table_alias is not None:
                    return replacement.rewrite(should_render_table_alias=should_render_table_alias)
                else:
                    return replacement
            else:
                if should_render_table_alias is not None:
                    return SqlColumnReferenceExpression(
                        col_ref=self.col_ref, should_render_table_alias=should_render_table_alias
                    )
                return self

        if should_render_table_alias is not None:
            return SqlColumnReferenceExpression(
                col_ref=self.col_ref, should_render_table_alias=should_render_table_alias
            )

        return SqlColumnReferenceExpression(
            col_ref=self.col_ref, should_render_table_alias=self.should_render_table_alias
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage(column_reference_exprs=(self,))

    @property
    def should_render_table_alias(self) -> bool:  # noqa: D
        return self._should_render_table_alias

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlColumnReferenceExpression):
            return False
        return self.col_ref == other.col_ref


class SqlColumnAliasReferenceExpression(SqlExpressionNode):
    """An expression that evaluates to the alias of a column, but is not qualified with a table alias.

    e.g. SELECT foo vs. SELECT a.foo.

    This is needed to handle some exceptional cases, but in general, this should not be used as it can lead to
    ambiguities.
    """

    def __init__(self, column_alias: str) -> None:  # noqa: D
        self._column_alias = column_alias
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_COLUMN_REFERENCE_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D
        return visitor.visit_column_alias_reference_expr(self)

    @property
    def column_alias(self) -> str:  # noqa: D
        return self._column_alias

    @property
    def description(self) -> str:  # noqa: D
        return f"Unqualified Column: {self._column_alias}"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [DisplayedProperty("column_alias", self.column_alias)]

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    @property
    def as_column_reference_expression(self) -> Optional[SqlColumnReferenceExpression]:  # noqa:
        return None

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        if column_replacements:
            raise NotImplementedError()
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage(column_alias_reference_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlColumnAliasReferenceExpression):
            return False
        return self.column_alias == other.column_alias


class SqlComparison(Enum):  # noqa: D
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUALS = "<="
    GREATER_THAN_OR_EQUALS = ">="
    EQUALS = "="


class SqlComparisonExpression(SqlExpressionNode):
    """A comparison using >, <, <=, >=, =.

    e.g. my_table.my_column = a + b
    """

    def __init__(self, left_expr: SqlExpressionNode, comparison: SqlComparison, right_expr: SqlExpressionNode) -> None:
        """Constructor.

        Args:
            left_expr: The expression on the left side of the =
            comparison: The comparison to use on expressions
            right_expr: The expression on the right side of the =
        """
        self._left_expr = left_expr
        self._comparison = comparison
        self._right_expr = right_expr
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[self._left_expr, self._right_expr])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_COMPARISON_ID_PREFIX

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_comparison_expr(self)

    @property
    def left_expr(self) -> SqlExpressionNode:  # noqa: D
        return self._left_expr

    @property
    def right_expr(self) -> SqlExpressionNode:  # noqa: D
        return self._right_expr

    @property
    def description(self) -> str:  # noqa: D
        return f"{self._comparison.value} Expression"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return super().displayed_properties + [
            DisplayedProperty("left_expr", self.left_expr),
            DisplayedProperty("comparison", self.comparison.value),
            DisplayedProperty("right_expr", self.right_expr),
        ]

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return True

    @property
    def comparison(self) -> SqlComparison:  # noqa: D
        return self._comparison

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlComparisonExpression(
            left_expr=self.left_expr.rewrite(column_replacements, should_render_table_alias),
            comparison=self.comparison,
            right_expr=self.right_expr.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlComparisonExpression):
            return False
        return self.comparison == other.comparison and self._parents_match(other)


class SqlFunction(Enum):
    """Names of known SQL functions like SUM() in SELECT SUM(...)

    Values are the SQL string to be used in rendering.
    """

    # Aggregation functions
    AVERAGE = "AVG"
    # Most engines implement count_distinct as a leading DISTINCT keyword like `COUNT(DISTINCT col1, col2...)`
    COUNT_DISTINCT = "COUNT"
    MAX = "MAX"
    MIN = "MIN"
    SUM = "SUM"

    # Field management functions
    COALESCE = "COALESCE"
    CONCAT = "CONCAT"

    @staticmethod
    def distinct_aggregation_functions() -> Sequence[SqlFunction]:
        """Returns a tuple containg all currently-supported DISTINCT type aggregation functions

        This is not a property because properties don't play nicely with static/class methods.
        """
        return (SqlFunction.COUNT_DISTINCT,)

    @staticmethod
    def is_distinct_aggregation(function_type: SqlFunction) -> bool:
        """Convenience method to check if the input function is a distinct aggregation type.

        This is useful in SQL expression rendering, as most engines implement distinct as a keyword modifier on
        an argument (e.g., `COUNT(DISTINCT expr)`) while our model handling and rendering supports distinct functions
        (e.g., `count_distinct(expr)`) and otherwise does not guarantee correct results when the DISTINCT keyword
        is used.
        """
        return function_type in SqlFunction.distinct_aggregation_functions()

    @staticmethod
    def is_aggregation(function_type: SqlFunction) -> bool:
        """Returns true if the given function is an aggregation function."""

        return function_type in (
            SqlFunction.AVERAGE,
            SqlFunction.COUNT_DISTINCT,
            SqlFunction.MAX,
            SqlFunction.MIN,
            SqlFunction.SUM,
        )

    @staticmethod
    def from_aggregation_type(aggregation_type: AggregationType) -> SqlFunction:
        """Converter method to get the SqlFunction value corresponding to the given AggregationType

        Make sure to leave the else: block in place, as this enforces an exhaustive switch through the
        AggregationType enumeration values.
        """
        if aggregation_type is AggregationType.AVERAGE:
            return SqlFunction.AVERAGE
        elif aggregation_type is AggregationType.COUNT_DISTINCT:
            return SqlFunction.COUNT_DISTINCT
        elif aggregation_type is AggregationType.MAX:
            return SqlFunction.MAX
        elif aggregation_type is AggregationType.MIN:
            return SqlFunction.MIN
        elif aggregation_type is AggregationType.SUM:
            return SqlFunction.SUM
        elif (
            aggregation_type is AggregationType.SUM_BOOLEAN
            or aggregation_type is AggregationType.BOOLEAN
            or aggregation_type is AggregationType.COUNT
        ):
            raise RuntimeError(
                f"Unhandled aggregation type {aggregation_type} - this should have been transformed to SUM "
                "during model parsing."
            )
        else:
            assert_values_exhausted(aggregation_type)


class SqlFunctionExpression(SqlExpressionNode):
    """A function expression like SUM(1)."""

    @staticmethod
    def from_aggregation_type(
        aggregation_type: AggregationType, sql_column_expression: SqlColumnReferenceExpression
    ) -> SqlFunctionExpression:
        """Given the aggregation type, return an SQL function expression that does that aggregation on the given col."""
        return SqlFunctionExpression(
            sql_function=SqlFunction.from_aggregation_type(aggregation_type=aggregation_type),
            sql_function_args=[sql_column_expression],
        )

    def __init__(self, sql_function: SqlFunction, sql_function_args: List[SqlExpressionNode]) -> None:
        """Constructor.

        Args:
            sql_function: The function that this represents.
            sql_function_args: The arguments that should go into the function. e.g. for "CONCAT(a, b)", the arg
            expressions should be "a" and "b".
        """
        self._sql_function = sql_function
        self._sql_function_args = sql_function_args
        super().__init__(node_id=self.create_unique_id(), parent_nodes=sql_function_args)

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_FUNCTION_ID_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_function_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return f"{self._sql_function.value} Expression"

    @property
    def displayed_properties(self) -> List[DisplayedProperty]:  # noqa: D
        return (
            super().displayed_properties
            + [DisplayedProperty("function", self.sql_function)]
            + [DisplayedProperty("argument", x) for x in self.sql_function_args]
        )

    @property
    def sql_function(self) -> SqlFunction:  # noqa: D
        return self._sql_function

    @property
    def sql_function_args(self) -> List[SqlExpressionNode]:  # noqa: D
        return self._sql_function_args

    def __repr__(self) -> str:  # noqa: D
        return f"{self.__class__.__name__}(node_id={self.node_id}, sql_function={self.sql_function.name})"

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlFunctionExpression(
            sql_function=self.sql_function,
            sql_function_args=[
                x.rewrite(column_replacements, should_render_table_alias) for x in self.sql_function_args
            ],
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(function_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlFunctionExpression):
            return False
        return self.sql_function == other.sql_function and self._parents_match(other)


class SqlNullExpression(SqlExpressionNode):
    """Represents NULL."""

    def __init__(self) -> None:  # noqa: D
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_null_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "NULL Expression"

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return self

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage(other_exprs=(self,))

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        return isinstance(other, SqlNullExpression)


class SqlLogicalOperator(Enum):
    """List all supported binary logical operator expressions.

    Value is the SQL string used when rendering the operator.
    """

    AND = "AND"
    OR = "OR"


class SqlLogicalExpression(SqlExpressionNode):
    """A logical expression like "a AND b AND c"."""

    def __init__(self, operator: SqlLogicalOperator, args: Tuple[SqlExpressionNode, ...]) -> None:  # noqa: D
        self._operator = operator
        super().__init__(node_id=self.create_unique_id(), parent_nodes=list(args))

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_LOGICAL_OPERATOR_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return True

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_logical_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return f"Logical Operator {self._operator.value}"

    @property
    def args(self) -> Sequence[SqlExpressionNode]:  # noqa: D
        return self.parent_nodes

    @property
    def operator(self) -> SqlLogicalOperator:  # noqa: D
        return self._operator

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlLogicalExpression(
            operator=self.operator,
            args=tuple(x.rewrite(column_replacements, should_render_table_alias) for x in self.args),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlLogicalExpression):
            return False
        return self.operator == other.operator and self._parents_match(other)


class SqlIsNullExpression(SqlExpressionNode):
    """An IS NULL expression like "foo IS NULL"."""

    def __init__(self, arg: SqlExpressionNode) -> None:  # noqa: D
        self._arg = arg
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[arg])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_IS_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return True

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_is_null_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "IS NULL Expression"

    @property
    def arg(self) -> SqlExpressionNode:  # noqa: D
        return self._arg

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlIsNullExpression(arg=self.arg.rewrite(column_replacements, should_render_table_alias))

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine([self.arg.lineage, SqlExpressionTreeLineage(other_exprs=(self,))])

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlIsNullExpression):
            return False
        return self._parents_match(other)


class SqlTimeDeltaExpression(SqlExpressionNode):
    """create time delta between eg `DATE_SUB(ds, 2, month)`"""

    def __init__(  # noqa: D
        self,
        arg: SqlExpressionNode,
        count: int,
        granularity: TimeGranularity,
        grain_to_date: Optional[TimeGranularity] = None,
    ) -> None:
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[arg])
        self._count = count
        self._time_granularity = granularity
        self._arg = arg
        self._grain_to_date = grain_to_date

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_IS_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_time_delta_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "Time delta"

    @property
    def arg(self) -> SqlExpressionNode:  # noqa: D
        return self._arg

    @property
    def grain_to_date(self) -> Optional[TimeGranularity]:  # noqa: D
        return self._grain_to_date

    @property
    def count(self) -> int:  # noqa: D
        return self._count

    @property
    def granularity(self) -> TimeGranularity:  # noqa: D
        return self._time_granularity

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlTimeDeltaExpression(
            arg=self.arg.rewrite(column_replacements, should_render_table_alias),
            count=self.count,
            granularity=self.granularity,
            grain_to_date=self.grain_to_date,
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlTimeDeltaExpression):
            return False
        return (
            self.count == other.count
            and self.granularity == other.granularity
            and self.grain_to_date == other.grain_to_date
            and self._parents_match(other)
        )


class SqlCastToTimestampExpression(SqlExpressionNode):
    """Cast to the timestamp type like CAST('2020-01-01' AS TIMESTAMP)"""

    def __init__(self, arg: SqlExpressionNode) -> None:  # noqa: D
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[arg])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_IS_NULL_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D
        return visitor.visit_cast_to_timestamp_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "Cast to Timestamp"

    @property
    def arg(self) -> SqlExpressionNode:  # noqa: D
        assert len(self.parent_nodes) == 1
        return self.parent_nodes[0]

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlCastToTimestampExpression(arg=self.arg.rewrite(column_replacements, should_render_table_alias))

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlCastToTimestampExpression):
            return False
        return self._parents_match(other)


class SqlDateTruncExpression(SqlExpressionNode):
    """Apply a date trunc to a column like CAST('2020-01-01' AS TIMESTAMP)"""

    def __init__(self, time_granularity: TimeGranularity, arg: SqlExpressionNode) -> None:
        """Constructor.

        Args:
            time_granularity: the granularity to DATE_TRUNC() to.
            arg: the value to DATE_TRUNC().
        """
        self._time_granularity = time_granularity
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[arg])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_DATE_TRUNC

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_date_trunc_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return f"DATE_TRUNC() to {self.time_granularity}"

    @property
    def time_granularity(self) -> TimeGranularity:  # noqa: D
        return self._time_granularity

    @property
    def arg(self) -> SqlExpressionNode:  # noqa: D
        assert len(self.parent_nodes) == 1
        return self.parent_nodes[0]

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlDateTruncExpression(
            time_granularity=self.time_granularity, arg=self.arg.rewrite(column_replacements, should_render_table_alias)
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlDateTruncExpression):
            return False
        return self.time_granularity == other.time_granularity and self._parents_match(other)


class SqlRatioComputationExpression(SqlExpressionNode):
    """Node for expressing Ratio metrics to allow for appropriate casting to float/double in each engine

    In future we might wish to break this up into a set of nodes, e.g., SqlCastExpression and SqlMathExpression
    or even add CAST to SqlFunctionExpression. However, at this time the only mathematical operation we encode
    is division, and we only use that for ratios. Similarly, the only times we do typecasting are when we are
    coercing timestamps (already handled) or computing ratio metrics.
    """

    def __init__(self, numerator: SqlExpressionNode, denominator: SqlExpressionNode) -> None:
        """Initialize this node for computing a ratio. Expression renderers should handle the casting

        Args:
            numerator: the expression for the numerator in the ratio
            denominator: the expression for the denominator in the ratio
        """
        self._numerator = numerator
        self._denominator = denominator
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[numerator, denominator])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_RATIO_COMPUTATION

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor) -> VisitorOutputT:  # noqa: D
        return visitor.visit_ratio_computation_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "Divide numerator by denominator, with appropriate casting"

    @property
    def numerator(self) -> SqlExpressionNode:  # noqa: D
        return self._numerator

    @property
    def denominator(self) -> SqlExpressionNode:  # noqa: D
        return self._denominator

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlRatioComputationExpression(
            numerator=self.numerator.rewrite(column_replacements, should_render_table_alias),
            denominator=self.denominator.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlRatioComputationExpression):
            return False
        return self._parents_match(other)


class SqlBetweenExpression(SqlExpressionNode):
    """A BETWEEN clause like `column BETWEEN val1 AND val2`"""

    def __init__(  # noqa: D
        self, column_arg: SqlExpressionNode, start_expr: SqlExpressionNode, end_expr: SqlExpressionNode
    ) -> None:
        self._column_arg = column_arg
        self._start_expr = start_expr
        self._end_expr = end_expr
        super().__init__(node_id=self.create_unique_id(), parent_nodes=[column_arg, start_expr, end_expr])

    @classmethod
    def id_prefix(cls) -> str:  # noqa: D
        return SQL_EXPR_BETWEEN_PREFIX

    @property
    def requires_parenthesis(self) -> bool:  # noqa: D
        return False

    def accept(self, visitor: SqlExpressionNodeVisitor[VisitorOutputT]) -> VisitorOutputT:  # noqa: D
        return visitor.visit_between_expr(self)

    @property
    def description(self) -> str:  # noqa: D
        return "BETWEEN operator"

    @property
    def column_arg(self) -> SqlExpressionNode:  # noqa: D
        return self._column_arg

    @property
    def start_expr(self) -> SqlExpressionNode:  # noqa: D
        return self._start_expr

    @property
    def end_expr(self) -> SqlExpressionNode:  # noqa: D
        return self._end_expr

    def rewrite(  # noqa: D
        self,
        column_replacements: Optional[SqlColumnReplacements] = None,
        should_render_table_alias: Optional[bool] = None,
    ) -> SqlExpressionNode:
        return SqlBetweenExpression(
            column_arg=self.column_arg.rewrite(column_replacements, should_render_table_alias),
            start_expr=self.start_expr.rewrite(column_replacements, should_render_table_alias),
            end_expr=self.end_expr.rewrite(column_replacements, should_render_table_alias),
        )

    @property
    def lineage(self) -> SqlExpressionTreeLineage:  # noqa: D
        return SqlExpressionTreeLineage.combine(
            tuple(x.lineage for x in self.parent_nodes) + (SqlExpressionTreeLineage(other_exprs=(self,)),)
        )

    def matches(self, other: SqlExpressionNode) -> bool:  # noqa: D
        if not isinstance(other, SqlBetweenExpression):
            return False
        return self._parents_match(other)



import pytest

from metricflow.dataclass_serialization import DataclassSerializer, DataClassDeserializer
from metricflow.instances import InstanceSet
from metricflow.test.fixtures.model_fixtures import ConsistentIdObjectRepository


@pytest.fixture
def serializer() -> DataclassSerializer:  # noqa: D
    return DataclassSerializer()


@pytest.fixture
def deserializer() -> DataClassDeserializer:  # noqa: D
    return DataClassDeserializer()


def test_serialization(  # noqa: D
    consistent_id_object_repository: ConsistentIdObjectRepository,
    serializer: DataclassSerializer,
    deserializer: DataClassDeserializer,
) -> None:
    for _, data_set in consistent_id_object_repository.simple_model_data_sets.items():
        serialized_obj = serializer.pydantic_serialize(data_set.instance_set)
        deserialized_obj = deserializer.pydantic_deserialize(dataclass_type=InstanceSet, serialized_obj=serialized_obj)
        assert data_set.instance_set == deserialized_obj



from __future__ import annotations
from typing import Optional, Tuple, Union

from metricflow.model.objects.base import FrozenBaseModel, PydanticCustomInputParser, PydanticParseableValueType


class SqlTable(PydanticCustomInputParser, FrozenBaseModel):
    """Represents a reference to a SQL table."""

    db_name: Optional[str] = None
    schema_name: str
    table_name: str

    @classmethod
    def _from_yaml_value(cls, input: PydanticParseableValueType) -> SqlTable:
        """Parses a SqlTable from string input found in a user-provided model specification

        Raises a ValueError on any non-string input, as all user-provided specifications of table identifiers
        should be strings conforming to the expectations defined in the from_string method.
        """
        if isinstance(input, str):
            return SqlTable.from_string(input)
        else:
            raise ValueError(
                f"SqlTable inputs from model configs are expected to always be of type string, but got type "
                f"{type(input)} with value: {input}"
            )

    @staticmethod
    def from_string(sql_str: str) -> SqlTable:  # noqa: D
        sql_str_split = sql_str.split(".")
        if len(sql_str_split) == 2:
            return SqlTable(schema_name=sql_str_split[0], table_name=sql_str_split[1])
        elif len(sql_str_split) == 3:
            return SqlTable(db_name=sql_str_split[0], schema_name=sql_str_split[1], table_name=sql_str_split[2])
        raise RuntimeError(
            f"Invalid input for a SQL table, expected form '<schema>.<table>' or '<db>.<schema>.<table>' "
            f"but got: {sql_str}"
        )

    @property
    def sql(self) -> str:
        """Return the snippet that can be used for use in SQL queries."""
        if self.db_name:
            return f"{self.db_name}.{self.schema_name}.{self.table_name}"
        return f"{self.schema_name}.{self.table_name}"

    @property
    def parts_tuple(self) -> Union[Tuple[str, str], Tuple[str, str, str]]:
        """Return a tuple of the sql table parts"""
        if self.db_name:
            return (self.db_name, self.schema_name, self.table_name)
        else:
            return (self.schema_name, self.table_name)



import logging

import pytest

from metricflow.dataflow.builder.node_data_set import DataflowPlanNodeOutputDataSetResolver
from metricflow.dataflow.dataflow_plan import ReadSqlSourceNode, FilterElementsNode, MetricTimeDimensionTransformNode
from metricflow.dataset.convert_data_source import DataSourceToDataSetConverter
from metricflow.instances import DataSourceReference
from metricflow.model.semantic_model import SemanticModel
from metricflow.object_utils import pformat_big_objects
from metricflow.plan_conversion.column_resolver import DefaultColumnAssociationResolver
from metricflow.plan_conversion.dataflow_to_sql import DataflowToSqlQueryPlanConverter
from metricflow.plan_conversion.sql_dataset import SqlDataSet
from metricflow.plan_conversion.time_spine import TimeSpineSource
from metricflow.protocols.sql_client import SqlClient
from metricflow.specs import TimeDimensionSpec, TimeDimensionReference
from metricflow.sql.optimizer.optimization_levels import SqlQueryOptimizationLevel
from metricflow.sql.render.sql_plan_renderer import SqlQueryPlanRenderer
from metricflow.time.time_granularity import TimeGranularity

logger = logging.getLogger(__name__)


@pytest.mark.skip("Example for developers.")
def test_view_sql_generated_at_a_node(
    sql_client: SqlClient,
    simple_semantic_model: SemanticModel,
    time_spine_source: TimeSpineSource,
) -> None:
    """Example that shows how to view generated SQL for nodes in a dataflow plan."""
    bookings_data_source = simple_semantic_model.data_source_semantics.get_by_reference(
        DataSourceReference(data_source_name="bookings_source")
    )
    assert bookings_data_source
    column_association_resolver = DefaultColumnAssociationResolver(
        semantic_model=simple_semantic_model,
    )
    to_data_set_converter = DataSourceToDataSetConverter(column_association_resolver)

    to_sql_plan_converter = DataflowToSqlQueryPlanConverter[SqlDataSet](
        column_association_resolver=DefaultColumnAssociationResolver(simple_semantic_model),
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )
    sql_renderer: SqlQueryPlanRenderer = sql_client.sql_engine_attributes.sql_query_plan_renderer
    node_output_resolver = DataflowPlanNodeOutputDataSetResolver[SqlDataSet](
        column_association_resolver=column_association_resolver,
        semantic_model=simple_semantic_model,
        time_spine_source=time_spine_source,
    )

    # Show SQL and spec set at a source node.
    bookings_source_data_set = to_data_set_converter.create_sql_source_data_set(bookings_data_source)
    read_source_node = ReadSqlSourceNode[SqlDataSet](bookings_source_data_set)
    sql_plan_at_read_node = to_sql_plan_converter.convert_to_sql_query_plan(
        sql_engine_attributes=sql_client.sql_engine_attributes,
        sql_query_plan_id="example_sql_plan",
        dataflow_plan_node=read_source_node,
        optimization_level=SqlQueryOptimizationLevel.O4,
    )
    sql_at_read_node = sql_renderer.render_sql_query_plan(sql_plan_at_read_node).sql
    spec_set_at_read_node = node_output_resolver.get_output_data_set(read_source_node).instance_set.spec_set
    logger.info(f"SQL generated at {read_source_node} is:\n\n{sql_at_read_node}")
    logger.info(f"Spec set at {read_source_node} is:\n\n{pformat_big_objects(spec_set_at_read_node)}")

    metric_time_node = MetricTimeDimensionTransformNode(
        parent_node=read_source_node,
        aggregation_time_dimension_reference=TimeDimensionReference(element_name="ds"),
    )

    # Show SQL and spec set at a filter node.
    filter_elements_node = FilterElementsNode(
        parent_node=metric_time_node,
        include_specs=(
            TimeDimensionSpec(element_name="metric_time", identifier_links=(), time_granularity=TimeGranularity.DAY),
        ),
    )
    sql_plan_at_filter_elements_node = to_sql_plan_converter.convert_to_sql_query_plan(
        sql_engine_attributes=sql_client.sql_engine_attributes,
        sql_query_plan_id="example_sql_plan",
        dataflow_plan_node=filter_elements_node,
        optimization_level=SqlQueryOptimizationLevel.O4,
    )
    sql_at_filter_elements_node = sql_renderer.render_sql_query_plan(sql_plan_at_filter_elements_node).sql
    spec_set_at_filter_elements_node = node_output_resolver.get_output_data_set(
        filter_elements_node
    ).instance_set.spec_set
    logger.info(f"SQL generated at {filter_elements_node} is:\n\n{sql_at_filter_elements_node}")
    logger.info(f"Spec set at {filter_elements_node} is:\n\n{pformat_big_objects(spec_set_at_filter_elements_node)}")



from __future__ import annotations

from collections import OrderedDict
from dataclasses import dataclass
from typing import Tuple, Sequence, List

from metricflow.dataset.dataset import DataSet
from metricflow.protocols.semantics import DataSourceSemanticsAccessor
from metricflow.specs import (
    DimensionSpec,
    TimeDimensionSpec,
    InstanceSpecSet,
    PartitionSpecSet,
)


@dataclass(frozen=True)
class PartitionDimensionJoinDescription:
    """Describes which partition dimensions should be joined between nodes."""

    start_node_dimension_spec: DimensionSpec
    node_to_join_dimension_spec: DimensionSpec


@dataclass(frozen=True)
class PartitionTimeDimensionJoinDescription:
    """Describes which partition time dimensions should be joined between nodes."""

    start_node_time_dimension_spec: TimeDimensionSpec
    node_to_join_time_dimension_spec: TimeDimensionSpec


class PartitionJoinResolver:
    """When joining data sets, this class helps to figure out the necessary partition specs to join on."""

    def __init__(self, data_source_semantics: DataSourceSemanticsAccessor) -> None:  # noqa: D
        self._data_source_semantics = data_source_semantics

    def _get_partitions(self, spec_set: InstanceSpecSet) -> PartitionSpecSet:
        """Returns the specs from the instance set that correspond to partition specs."""
        partition_dimension_specs = tuple(
            x
            for x in spec_set.dimension_specs
            if self._data_source_semantics.get_dimension(dimension_reference=x.reference).is_partition
        )
        partition_time_dimension_specs = tuple(
            x
            for x in spec_set.time_dimension_specs
            if x.reference != DataSet.metric_time_dimension_reference()
            and self._data_source_semantics.get_time_dimension(time_dimension_reference=x.reference).is_partition
        )

        return PartitionSpecSet(
            dimension_specs=partition_dimension_specs,
            time_dimension_specs=partition_time_dimension_specs,
        )

    @staticmethod
    def _get_simplest_dimension_spec(dimension_specs: Sequence[DimensionSpec]) -> DimensionSpec:
        """Return the time dimension spec with the fewest identifier links."""
        assert len(dimension_specs) > 0
        sorted_dimension_specs = sorted(dimension_specs, key=lambda x: len(x.identifier_links))
        return sorted_dimension_specs[0]

    def resolve_partition_dimension_joins(
        self, start_node_spec_set: InstanceSpecSet, node_to_join_spec_set: InstanceSpecSet
    ) -> Tuple[PartitionDimensionJoinDescription, ...]:
        """Figures out which partition dimensions to join on."""
        start_node_partitions = self._get_partitions(start_node_spec_set)
        join_node_partitions = self._get_partitions(node_to_join_spec_set)

        partition_join_descriptions = []

        start_node_dimension_element_names = tuple(
            OrderedDict.fromkeys(x.element_name for x in start_node_partitions.dimension_specs)
        )

        for element_name in start_node_dimension_element_names:
            start_node_spec = PartitionJoinResolver._get_simplest_dimension_spec(
                tuple(x for x in start_node_partitions.dimension_specs if x.element_name == element_name)
            )
            join_node_time_dimension_element_names = tuple(
                OrderedDict.fromkeys(x.element_name for x in join_node_partitions.dimension_specs)
            )
            if element_name in join_node_time_dimension_element_names:
                join_node_spec = PartitionJoinResolver._get_simplest_dimension_spec(
                    tuple(x for x in join_node_partitions.dimension_specs if x.element_name == element_name)
                )
                partition_join_descriptions.append(
                    PartitionDimensionJoinDescription(
                        start_node_dimension_spec=start_node_spec,
                        node_to_join_dimension_spec=join_node_spec,
                    )
                )

        return tuple(partition_join_descriptions)

    @staticmethod
    def _get_simplest_time_dimension_spec(time_dimension_specs: Sequence[TimeDimensionSpec]) -> TimeDimensionSpec:
        """Return the time dimension spec with the smallest granularity, then fewest identifier links."""
        assert len(time_dimension_specs) > 0
        sorted_specs = sorted(time_dimension_specs, key=lambda x: (x.time_granularity, len(x.identifier_links)))
        return sorted_specs[0]

    def resolve_partition_time_dimension_joins(
        self, start_node_spec_set: InstanceSpecSet, node_to_join_spec_set: InstanceSpecSet
    ) -> Tuple[PartitionTimeDimensionJoinDescription, ...]:
        """Figures out which partition time dimensions to join on."""
        start_node_partitions = self._get_partitions(start_node_spec_set)
        join_node_partitions = self._get_partitions(node_to_join_spec_set)
        partition_join_descriptions: List[PartitionTimeDimensionJoinDescription] = []

        # Get all unique element names
        time_dimension_element_names = tuple(
            OrderedDict.fromkeys(x.element_name for x in start_node_partitions.time_dimension_specs)
        )
        for element_name in time_dimension_element_names:
            start_node_spec = PartitionJoinResolver._get_simplest_time_dimension_spec(
                tuple(x for x in start_node_partitions.time_dimension_specs if x.element_name == element_name)
            )
            join_node_time_dimension_element_names = tuple(
                OrderedDict.fromkeys(x.element_name for x in join_node_partitions.time_dimension_specs)
            )
            if element_name in join_node_time_dimension_element_names:
                join_node_spec = PartitionJoinResolver._get_simplest_time_dimension_spec(
                    tuple(x for x in join_node_partitions.time_dimension_specs if x.element_name == element_name)
                )
                partition_join_descriptions.append(
                    PartitionTimeDimensionJoinDescription(
                        start_node_time_dimension_spec=start_node_spec,
                        node_to_join_time_dimension_spec=join_node_spec,
                    )
                )
        return tuple(partition_join_descriptions)



import logging

from metricflow.aggregation_properties import AggregationType
from metricflow.model.objects.user_configured_model import UserConfiguredModel
from metricflow.model.transformations.transform_rule import ModelTransformRule

logger = logging.getLogger(__name__)


class BooleanMeasureAggregationRule(ModelTransformRule):
    """Converts the expression used in boolean measures so that it can be aggregated."""

    @staticmethod
    def transform_model(model: UserConfiguredModel) -> UserConfiguredModel:  # noqa: D
        for data_source in model.data_sources:
            for measure in data_source.measures:
                if measure.agg == AggregationType.BOOLEAN:
                    logger.warning(
                        f"In data source {data_source.name}, measure `{measure.reference.element_name}` "
                        f"is configured as aggregation type `boolean`, which has been deprecated. Please use "
                        f"`sum_boolean` instead."
                    )
                if measure.agg == AggregationType.BOOLEAN or measure.agg == AggregationType.SUM_BOOLEAN:
                    if measure.expr:
                        measure.expr = f"CASE WHEN {measure.expr} THEN 1 ELSE 0 END"
                    else:
                        measure.expr = f"CASE WHEN {measure.name} THEN 1 ELSE 0 END"

                    measure.agg = AggregationType.SUM

        return model



from __future__ import annotations

from typing import Optional, List
from metricflow.aggregation_properties import AggregationType
from metricflow.model.objects.common import Metadata
from metricflow.model.objects.base import ModelWithMetadataParsing, HashableBaseModel
from metricflow.references import MeasureReference, TimeDimensionReference


class NonAdditiveDimensionParameters(HashableBaseModel):
    """Describes the params for specifying non-additive dimensions in a measure.

    NOTE: Currently, only TimeDimensions are supported for this filter
    """

    name: str

    # Optional Fields
    window_choice: AggregationType = AggregationType.MIN
    window_groupings: List[str] = []


class Measure(HashableBaseModel, ModelWithMetadataParsing):
    """Describes a measure"""

    name: str
    agg: AggregationType
    description: Optional[str]
    create_metric: Optional[bool]
    expr: Optional[str] = None
    metadata: Optional[Metadata]
    non_additive_dimension: Optional[NonAdditiveDimensionParameters] = None

    # Defines the time dimension to aggregate this measure by. If not specified, it means to use the primary time
    # dimension in the data source.
    agg_time_dimension: Optional[str] = None

    @property
    def checked_agg_time_dimension(self) -> TimeDimensionReference:
        """Returns the aggregation time dimension, throwing an exception if it's not set."""
        assert self.agg_time_dimension, (
            f"Aggregation time dimension for measure {self.name} is not set! This should either be set directly on "
            f"the measure specification in the model, or else defaulted to the primary time dimension in the data "
            f"source containing the measure."
        )
        return TimeDimensionReference(element_name=self.agg_time_dimension)

    @property
    def reference(self) -> MeasureReference:  # noqa: D
        return MeasureReference(element_name=self.name)
